import numpy as np
import pandas as pd

__all__ = ['woe', 'iv']


def woe(feature, y_true, pos_label=1):
    """Calculate series woe value
    
    Args:
        feature: pd.Series, shape (n_samples,) x variable, model feature.
        y_true: pd.Series, shape (n_samples,) The target variable for supervised learning problems.
        pos_label: int, default=1, positive label value.
    
    Return:
        series woe value
    """
    t = pd.DataFrame({'label':y_true, 'feature':feature})
    assert t.label.nunique()==2, "`y_true` should be binary classification."
    label_dict = {i:1 if i==pos_label else 0 for i in t.label.unique()}
    t['label'] = t.label.replace(label_dict)
    corr = t.label.sum()/(t.label.count()-t.label.sum())
    t = t.groupby(['feature']).label.apply(lambda x:np.log(x.sum()/(x.count()-x.sum())/corr))
    return t


def iv(feature, y_true, pos_label=1):
    """Calculate series iv value
    
    Args:
        feature: pd.Series, shape (n_samples,), x variable, model feature.
        y_true: pd.Series, shape (n_samples,), The target variable for supervised learning problems.
        pos_label: int, default=1, positive label value.
    
    Return:
        series iv value
    """
    t = pd.DataFrame({'label':y_true, 'feature':feature})
    assert t.label.nunique()==2, "`y_true` should be binary classification."
    label_dict = {i:1 if i==pos_label else 0 for i in t.label.unique()}
    t['label'] = t.label.replace(label_dict)
    def _iv(group, df):
        a = group.sum()/df.label.sum()
        b = (group.count()-group.sum())/(t.label.count()-t.label.sum())
        return (a-b)/np.log(a/b)
    corr = t.label.sum()/(t.label.count()-t.label.sum())
    t = t.groupby(['feature']).label.apply(lambda x:_iv(x, t)).sum()
    return t



import gc
import io
import warnings

import av
import numpy as np

__all__ = ['read_vedio', 'read_vedio_timestamps', 'save_vedio', 'VedioStream']



# def format_convert(filename_in, filename_out):
#     """transform vedio file format.
    
#     Args
#         filename_in: input vedio path.
#         filename_out: output vedio path.
#     """
#     input_ = av.open(filename_in)
#     output = av.open(filename_out, "w")

#     in_stream = input_.streams.video[0]
#     out_stream = output.add_stream(template=in_stream)
#     for packet in input_.demux(in_stream):
#         if packet.dts is None:
#             continue
#         packet.stream = out_stream
#         output.mux(packet)
    
#     in_audio = input_.streams.audio[0]
#     out_audio = output.add_stream(template=in_audio)
#     for packet in input_.demux(in_audio):
#         if packet.dts is None:
#             continue
#         packet.stream = out_audio
#         output.mux(packet)

#     input_.close()
#     output.close()

# def format_convert(filename_in, filename_out):
#     """transform vedio file format.
    
#     Args
#         filename_in: input vedio path.
#         filename_out: output vedio path.
#     """
#     input_ = av.open(filename_in)
#     output = av.open(filename_out, "w")

#     in_stream = input_.streams.video[0]
#     out_stream = output.add_stream(template=in_stream)

#     for packet in input_.demux(in_stream):
#         if packet.dts is None:
#             continue
#         packet.stream = out_stream
#         output.mux(packet)

#     input_.close()
#     output.close()
        
        
def save_vedio(filename, video_array, video_fps, video_codec="libx264", options=None,
               audio_array=None, audio_fps=None, audio_codec=None, audio_options=None):
    """Writes a 4d array in [H, W, C, N] format in a video file
    
    Args:
        filename: path where the video will be saved
        video_array: array containing the individual frames, as a uint8 array in [H, W, C, N] format,
            or array of list or Image of list.
        video_fps: video frames per second
        video_codec: the name of the video codec, i.e. "libx264", "h264", etc.
        options: dictionary containing options to be passed into the PyAV video stream
        audio_array: array[C, N] containing the audio, where C is the number of channels
            and N is the number of samples
        audio_fps: audio sample rate, typically 44100 or 48000
        audio_codec: the name of the audio codec, i.e. "mp3", "aac", etc.
        audio_options: dictionary containing options to be passed into the PyAV audio stream
    """
    container = av.open(filename, mode="w")
    stream = container.add_stream(video_codec, rate=int(video_fps))
    stream.pix_fmt = "yuv420p" if video_codec != "libx264rgb" else "rgb24"
    stream.options = options or {}
    if isinstance(vedio_array, np.ndarray):
        assert vedio_array.ndim==4, '`video_array` type error.'
        stream.width = video_array.shape[1]
        stream.height = video_array.shape[0]
        for i in range(vedio_array.shape[-1]):
            frame = av.VideoFrame.from_ndarray(vedio_array[:,:,:,i], format="rgb24")
            frame.pict_type = "NONE"
            for packet in stream.encode(frame):
                container.mux(packet)
    else:
        if isinstance(vedio_array, list):
            if isinstance(vedio_array[0], np.ndarray):
                assert vedio_array[0].ndim==3, '`video_array` type error.'
                stream.width = video_array[0].shape[1]
                stream.height = video_array[0].shape[0]
                for i in vedio_array:
                    frame = av.VideoFrame.from_ndarray(i, format="rgb24")
                    frame.pict_type = "NONE"
                    for packet in stream.encode(frame):
                        container.mux(packet)
            else:
                try:
                    stream.width = video_array[0].width
                    stream.height = video_array[0].height
                    for i in vedio_array:
                        frame = av.VideoFrame.from_image(i)
                        frame.pict_type = "NONE"
                        for packet in stream.encode(frame):
                            container.mux(packet)
                except:
                    raise ValueError('`video_array` type error.')
        else:
            raise ValueError('`video_array` type error.')
    for packet in stream.encode():
        container.mux(packet)

    if audio_array is not None:
        audio_format_dtypes = {
            "dbl": "<f8",
            "dblp": "<f8",
            "flt": "<f4",
            "fltp": "<f4",
            "s16": "<i2",
            "s16p": "<i2",
            "s32": "<i4",
            "s32p": "<i4",
            "u8": "u1",
            "u8p": "u1",
        }
        a_stream = container.add_stream(audio_codec, rate=int(audio_fps))
        a_stream.options = audio_options or {}

        num_channels = audio_array.shape[0]
        audio_layout = "stereo" if num_channels > 1 else "mono"
        audio_sample_fmt = container.streams.audio[0].format.name

        dtype = np.dtype(audio_format_dtypes[audio_sample_fmt])

        frame = av.AudioFrame.from_ndarray(audio_array.astype(dtype), format=audio_sample_fmt, layout=audio_layout)
        frame.sample_rate = int(audio_fps)

        for packet in a_stream.encode(frame):
            container.mux(packet)

        for packet in a_stream.encode():
            container.mux(packet)
    container.close()

        
def read_vedio(src, start_sec=0, end_sec=np.inf, vedio_format="HWCN", vedio_type=np.uint8, audio_format='CL'):
    """Reads a video from a file, returning both the video frames and the audio frames.
    
    Args:
        src: path to the video file or bytes.
        start_sec: float, The start presentation time of the video.
        end_sec: float, The end presentation time of the video.
        vedio_format: The format of the output video shape. Can be "HWCN" (default), "Image" return pillow instance.
        vedio_type: The dtype of the output video.
        audio_format: The format of the output audio shape. 'CL' or 'LC' e.g. 
    Returns:
        vedio: array[H, W, C, N]: the `N` video frames.
        audio: array[C, L]): the audio frames, where `C` is the number of channels and `L` is the number of points.
        metadata: metadata for the video and audio.
    """
    vedio_format = vedio_format.upper()
    if vedio_format!='IMAGE':
        if len([i for i in set(vedio_format) if i in 'HWCN'])!=len(vedio_format):
            raise ValueError(f"`vedio_format` should be 'HWCN' or 'NCHW' e.g., got {vedio_format}.")

    if end_sec < start_sec:
        raise ValueError("end_sec should be larger than start_sec")

    if isinstance(src, bytes):
        src = io.BytesIO(src)
        info = {'filename':''}
    else:
        info = {'filename':src}

    try:
        with av.open(src, metadata_errors="ignore") as container:
            if container.streams.video:
                container.streams.video[0].thread_type = "AUTO" 
                video_fps = container.streams.video[0].average_rate
                if video_fps is not None:
                    info["video_fps"]  = float(video_fps)
                info["vedio_duration"] = container.streams.video[0].duration*container.streams.video[0].time_base
                info["vedio_bitrate"]  = container.bit_rate
                info["vedio_frames"]   = container.streams.video[0].frames
                info["vedio_shape"]    = (container.streams.video[0].codec_context.height, 
                                          container.streams.video[0].codec_context.width)
                
                vframes = [frame for frame in container.decode(video=0) if start_sec<float(frame.pts*frame.time_base)<end_sec]
                if vedio_format!='IMAGE':
                    if vframes:
                        vframes = [frame.to_rgb().to_ndarray() for frame in vframes]
                        vframes = np.stack(vframes).astype(vedio_type)
                    else:
                        vframes = np.empty((0, 1, 1, 3), dtype=np.uint8)
                    if vedio_format!='NHWC':
                        transpose = {'N':0, 'H':1, 'W':2, 'C':3}
                        vframes = vframes.transpose(tuple(transpose[i] for i in vedio_format))
                else:
                    vframes = [frame.to_image() for frame in vframes]
        with av.open(src, metadata_errors="ignore") as container:
            if container.streams.audio:
                info["audio_fps"]       = container.streams.audio[0].rate
                info["audio_channel"]   = container.streams.audio[0].codec_context.channels
                info["audio_frames"]    = container.streams.audio[0].duration
                info["audio_duration"]  = container.streams.audio[0].duration*container.streams.audio[0].time_base
                
                aframes = [frame.to_ndarray() for frame in container.decode(audio=0) if start_sec<float(frame.pts*frame.time_base)<end_sec]
                if aframes:
                    aframes = np.concatenate(aframes, axis=1, dtype=np.float32)
                    if aframes.shape[0]!=info["audio_channel"]:
                        aframes = aframes.reshape(-1, info["audio_channel"]).T
                else:
                    aframes = np.empty((1, 0), dtype=np.float32)
                if audio_format!='CL':
                    transpose = {'C':0, 'L':1}
                    aframes = aframes.transpose(tuple(transpose[i] for i in audio_format))
    except av.AVError:
        pass
    gc.collect()
    return {'vedio':vframes, 'audio':aframes, 'metadata':info}


def read_vedio_timestamps(filename, pts_unit="sec"):
    """List the video frames timestamps.
    Note that the function decodes the whole video frame-by-frame.
    
    Args:
        filename: path to the video file
        pts_unit: unit in which timestamp values will be returned
            either 'pts' or 'sec'. Defaults to 'pts'.
    Returns:
        pts (List[int] if pts_unit = 'pts', List[Fraction] if pts_unit = 'sec'):
            presentation timestamps for each one of the frames in the video.
        video_fps (float, optional): the frame rate for the video
    """
    video_fps = None
    pts = []

    try:
        container = av.open(filename, metadata_errors="ignore")
        if container.streams.video:
            video_stream = container.streams.video[0]
            video_time_base = video_stream.time_base
            try:
                extradata = container.streams[0].codec_context.extradata
                if extradata is None:
                    pts = [x.pts for x in container.decode(video=0) if x.pts is not None]
                elif b"Lavc" in extradata:
                    pts = [x.pts for x in container.demux(video=0) if x.pts is not None]
                else:
                    pts = [x.pts for x in container.decode(video=0) if x.pts is not None]
            except av.AVError:
                warnings.warn(f"Failed decoding frames for file {filename}")
            video_fps = float(video_stream.average_rate)
        container.close()
    except av.AVError as e:
        msg = f"Failed to open container for {filename}; Caught error: {e}"
        warnings.warn(msg, RuntimeWarning)

    pts.sort()
    if pts_unit == "sec":
        pts = [x * video_time_base for x in pts]

    return {'timestamps':pts, 'video_fps':video_fps}


class VedioStream():
    def __init__(self, src, batch=1, data_format="HWCN", start_sec=0, end_sec=np.inf, dtype=np.uint8):
        """Converts a Vedio instance to a Numpy array.
    
        - if data_format is 'CL' or 'LC' e.g. return audio array.
          batch=4096, dtype=np.float32
        
        Args:
            src: input vedio path or bytes.
            batch: each iter batch number.
            data_format: array data format, eg.'HWCN', 'CHWN'. 
                'image' return pillow instance.
                'CL' or 'LC' e.g. return audio array.
            start_sec: float, The start presentation time of the video.
            end_sec: float, The end presentation time of the video.
            dtype: Dtype to use for the returned array.
                if 'CL' or 'LC' e.g. dtype is np.float32
        Returns:
            A Numpy array iterator.
        """
        self._batch = int(batch)
        self._data_format = data_format.upper()
        if 'L' in self._data_format:
            if len([i for i in set(self._data_format) if i in 'CL'])!=len(self._data_format):
                raise ValueError(f"`data_format` should be 'CL' or 'LC' e.g., got {data_format}.")
            self._dtype = np.float32
            if int(batch)==1:
                self._batch = 4096
        elif self._data_format!='IMAGE':
            if len([i for i in set(self._data_format) if i in 'HWCN'])!=len(self._data_format):
                raise ValueError(f"`data_format` should be 'HWCN' or 'NCHW' e.g., got {data_format}.")
            self._dtype = dtype
            
        if end_sec < start_sec:
            raise ValueError("end_sec should be larger than start_sec")
        self._end_sec = end_sec
        self._start_sec = start_sec
        
        if isinstance(src, bytes):
            src = io.BytesIO(src)
            self.metadata = {'filename':''}
        else:
            self.metadata = {'filename':src}
            
        self._container = av.open(src, metadata_errors="ignore")
        try:
            if self._container.streams.video:
                video_fps = self._container.streams.video[0].average_rate
                if video_fps is not None:
                    self.metadata["vedio_fps"]  = float(video_fps)
                self.metadata["vedio_duration"] = float(self._container.streams.video[0].duration*self._container.streams.video[0].time_base)
                self.metadata["vedio_bitrate"]  = self._container.bit_rate
                self.metadata["vedio_frames"]   = self._container.streams.video[0].frames
                self.metadata["vedio_shape"]    = (self._container.streams.video[0].codec_context.height, 
                                                   self._container.streams.video[0].codec_context.width)
            if self._container.streams.audio:
                self.metadata["audio_fps"]       = self._container.streams.audio[0].rate
                self.metadata["audio_channel"]   = self._container.streams.audio[0].codec_context.channels
                self.metadata["audio_frames"]    = self._container.streams.audio[0].duration
                self.metadata["audio_duration"]  = float(self.metadata["audio_frames"]*self._container.streams.audio[0].time_base)
        except av.AVError:
            pass
        if 'L' in self._data_format:
            try:
                self._aframes = np.concatenate([frame.to_ndarray() for frame in self._container.decode(audio=0)],
                                               axis=1, dtype=np.float32)
                if self._aframes.shape[0]!=self.metadata["audio_channel"]:
                    self._aframes = self._aframes.reshape(-1, self.metadata["audio_channel"]).T
                mix = self._container.streams.audio[0].start_time/1000000
                interval = float(self.metadata["audio_duration"])-mix
                self._batch_start_time = max(self._start_sec, mix)
                self._batch_end_time = min(self._end_sec, float(self.metadata["audio_duration"]))
                self._batch_time = self._batch/self._aframes.shape[1]*interval
                start = (self._batch_start_time-mix)/interval
                end = (self._batch_end_time-mix)/interval
                self._aframes = self._aframes[:, int(self._aframes.shape[1]*start):int(self._aframes.shape[1]*end)]
            except av.AVError:
                self._aframes = np.empty((1, 0), dtype=np.float32)
        else:
            self._c = self._container.decode(video=0)

    def __next__(self):
        try:
            batch_array = []
            batch_pts = []
            count = 0
            if 'L' in self._data_format:
                if not self._aframes.shape[1]:
                    self._container.close()
                    raise StopIteration
                batch_array = self._aframes[:,:self._batch]
                self._aframes = self._aframes[:,self._batch:]
                batch_pts.append(self._batch_start_time)
                if batch_array.shape[1]<self._batch:
                    batch_pts.append(self._batch_end_time)
                else:
                    self._batch_start_time += self._batch_time
                    batch_pts.append(self._batch_start_time)
                if self._data_format!='CL':
                    transpose = {'C':0, 'L':1}
                    batch_array = batch_array.transpose(tuple(transpose[i] for i in self._data_format))
            elif self._data_format=='IMAGE':
                try: 
                    while 1:
                        frame = next(self._c)
                        time_stamp = float(frame.pts*frame.time_base)
                        if self._start_sec<time_stamp<self._end_sec:                            
                            batch_array.append(frame.to_image())
                            batch_pts.append(time_stamp)
                            count += 1
                        if count==self._batch:
                            break
                except:
                    pass
                if not batch_array:
                    self._container.close()
                    raise StopIteration
            else:
                while 1:
                    try:
                        frame = next(self._c)
                        time_stamp = float(frame.pts*frame.time_base)
                        if self._start_sec<time_stamp<self._end_sec:
                            batch_array.append(frame.to_rgb().to_ndarray())
                            batch_pts.append(time_stamp)
                            count += 1
                        if count==self._batch:
                            break
                    except:
                        break
                if not batch_array:
                    self._container.close()
                    raise StopIteration
                batch_array = np.stack(batch_array).astype(self._dtype)
                if self._data_format!='NHWC':
                    transpose = {'N':0, 'H':1, 'W':2, 'C':3}
                    batch_array = batch_array.transpose(tuple(transpose[i] for i in self._data_format))
        except av.error.EOFError:
            self._container.close()
            raise StopIteration
        return {"data": batch_array, "pts": batch_pts}

    def __iter__(self):
        return self
    





from collections import defaultdict

import numpy as np
import pandas as pd

from linora.utils._config import Config
from linora.utils.pip._pip import install

__all__ = ['MysqlIoDataset', 'HiveIoDataset']


class DataSet():
    def __init__(self):
        self._params_init()
    
    def _params_init(self):
        self._params = Config()
        self._params.batch = 0
        self._params.batch_size = 1
        self._params.take_size = -1
        self._params.shuffle_size = 1
        self._params.prefetch_size = 1
        self._params.repeat_size = 1
        self._params.sample = 0
        self._params.step = 1
        self._params.tensor_mode = 'numpy'
        self._params.options = defaultdict(dict)
        
    def batch(self, batch_size, drop_remainder=False):
        """Combines consecutive elements of this dataset into batches.
        
        Args:
            batch_size: representing the number of consecutive elements of this dataset to combine in a single batch.
            drop_remainder: representing whether the last batch should be dropped in the case it has fewer than batch_size elements; 
                            the default behavior is not to drop the smaller batch.
        """
        assert 'batch' not in self._params.options, '`batch` already exists.'
        assert isinstance(batch_size, int) and batch_size>0, '`batch_size` type should be int and greater than 0.'
        self._params.batch_size = batch_size
        self._params.drop_remainder = drop_remainder
        self._params.options['batch'].update({self._params.step: {'batch_size':batch_size, 'drop_remainder':drop_remainder}})
        self._params.step += 1
        return self
        
    def enumerate(self, start=0):
        """Enumerates the elements of this dataset.
        
        Args:
            start: int, representing the start value for enumeration.
        """
        assert 'enumerate' not in self._params.options, '`enumerate` already exists.'
        self._params.enumerate = start
        self._params.options['enumerate'].update({self._params.step: {'start':start}})
        self._params.step += 1
        return self
    
    def map(self, map_func):
        """Maps map_func across the elements of this dataset.
        
        Args:
            map_func: A function mapping a dataset element to another dataset element.
            map_size: representing the number elements to process asynchronously in parallel. 
        """
        assert 'map' not in self._params.options, '`map` already exists.'
        self._params.map_func = map_func
        self._params.options['map'].update({self._params.step: {'map_func':map_func}})
        self._params.step += 1
        return self
    
    def options(self):
        """Returns the options for this dataset and its inputs."""
        return self._params.options
    
    def prefetch(self, prefetch_size):
        """Creates a Dataset that prefetches elements from this dataset.
        
        Args:
            prefetch_size: representing the maximum number of elements that will be buffered when prefetching.
        """
        assert 'prefetch' not in self._params.options, '`prefetch` already exists.'
        assert isinstance(prefetch_size, int) and prefetch_size>0, '`prefetch_size` type should be int and greater than 0.'
        self._params.prefetch_size = prefetch_size
        self._params.options['prefetch'].update({self._params.step: {'prefetch_size':prefetch_size}})
        self._params.step += 1
        return self

    def read_sql(self, sql, limit_size):
        """Read sql form database.
        
        Args:
            sql: query database content.
            limit_size: int, read data amount.
        """
        self._params.sql = sql
        if 'limit' not in self._params.sql:
            self._params.sql = f"{self._params.sql} limit {limit_size}"
        self._params.options['read_sql'].update({self._params.step: {'read_sql':sql}})
        self._params.step += 1
        return self
        
    def repeat(self, repeat_size):
        """Repeats this dataset so each original value is seen count times.
        
        Args:
            repeat_size: representing the number of times the dataset should be repeated.
        """
        assert isinstance(repeat_size, int) and repeat_size>0, '`repeat_size` type should be int and greater than 0.'
        self._params.repeat_size = repeat_size
        self._params.options['repeat'].update({self._params.step: {'repeat_size':repeat_size}})
        self._params.step += 1
        return self
    
    def shuffle(self, shuffle_size, seed=None):
        """Randomly shuffles the elements of this dataset.
        
        Args:
            shuffle_size: representing the number of elements from this dataset from which the new dataset will sample.
            seed: representing the random seed that will be used to create the distribution.
        """
        assert 'shuffle' not in self._params.options, '`shuffle` already exists.'
        assert isinstance(shuffle_size, int) and shuffle_size>-2 and shuffle_size!=0, '`shuffle_size` type should be int and greater than 0 or equal to -1.'
        self._params.shuffle_size = shuffle_size
        self._params.options['shuffle'].update({self._params.step: {'shuffle_size':shuffle_size, 'seed':seed}})
        self._params.step += 1
        return self
        
    def take(self, take_size):
        """Creates a Dataset with at most count elements from this dataset.
        
        Args:
            take_size: representing the number of elements of this dataset that should be taken to form the new dataset. 
                       If count is -1, or if count is greater than the size of this dataset, 
                       the new dataset will contain all elements of this dataset.
        """
        assert 'take' not in self._params.options, '`take` already exists.'
        assert isinstance(take_size, int) and take_size>-2 and take_size!=0, '`take_size` type should be int and greater than 0 or equal to -1.'
        self._params.take_size = take_size
        self._params.options['take'].update({self._params.step: {'take_size':take_size}})
        self._params.step += 1
        return self

    def to_tensor(self, mode='tf'):
        """Transform data from numpy array to tensor.
        
        Args:
            mode: Deep learning framework name, one of ['tf', 'pytorch', 'paddle', 'mxnet', 'mindspore'].
        """
        assert 'to_tensor' not in self._params.options, '`to_tensor` already exists.'
        assert 'take_while' not in self._params.options, '`take` must be placed in `take_while` front.'
        if mode in ['tf', 'tensorflow']:
            from tensorflow import convert_to_tensor
            self._params.framework = convert_to_tensor
        elif mode in ['pytorch', 'torch']:
            from torch import as_tensor
            self._params.framework = as_tensor
        elif mode in ['paddle', 'paddlepaddle']:
            from paddle import to_tensor
            self._params.framework = to_tensor
        elif mode in ['mx', 'mxnet']:
            from mxnet.ndarray import array
            self._params.framework = array
        elif mode in ['mindspore']:
            from mindspore.numpy import array
            self._params.framework = array
        else:
            raise ValueError('`mode` value error.')
        self._params.tensor_mode = mode
        self._params.options['to_tensor'].update({self._params.step: {'mode':mode}})
        self._params.step += 1
        return self
    
    def _to_tensor(self, data):
        if self._params.tensor_mode=='numpy':
            return data
        return self._params.framework(data)
    
    def __iter__(self):
        self._params.shuffle_size = int(np.ceil(max(self._params.shuffle_size, self._params.prefetch_size, 1)/self._params.batch_size)*self._params.batch_size)
        self._params.df = pd.read_sql(self._params.sql, self._params.engine, chunksize=self._params.shuffle_size)
        self._params.values = next(self._params.df)
        if 'shuffle' in self._params.options:
            self._params.values = self._params.values.sample(frac=1, random_state=self._params.shuffle_seed).reset_index(drop=True)
        self._params.repeat_size -= 1
        self._params.batch_index = 0
        return self
        
    def __next__(self):
        values = self._params.values.loc[self._params.batch_size*self._params.batch_index:self._params.batch_size*(self._params.batch_index+1)]
        if len(values)<self._params.batch_size:
            self._params.batch_index = 0
            try:
                self._params.values = next(self._params.df).reset_index(drop=True)
            except StopIteration:
                if self._params.repeat_size==0:
                    raise StopIteration
                self._params.df = pd.read_sql(self._params.sql, self._params.engine, chunksize=self._params.shuffle_size)
                self._params.values = next(self._params.df)
                self._params.repeat_size -= 1
            if 'shuffle' in self._params.options:
                self._params.values = self._params.values.sample(frac=1, random_state=self._params.shuffle_seed).reset_index(drop=True)
            values = self._params.values.loc[0:self._params.batch_size]
            
        self._params.batch += 1
        self._params.batch_index += 1
        if self._params.take_size>0:
            if self._params.sample>=self._params.take_size:
                raise StopIteration
            self._params.sample += len(values)
        if 'map' in self._params.options:
            return self._to_tensor(values.apply(self._params.map_func, axis=1).values)
        return self._to_tensor(values.values)


class MysqlIoDataset(DataSet):
    """Create a data set composed of mysql database.

    Args:
        host: mysql database host.
        port: mysql database port.
        user: mysql database user.
        password: mysql database password.
    """
    def __init__(self, host, port=3306, user=None, password=None, **kwargs):
        super(MysqlIoDataset, self).__init__()
        try:
            import pymysql
        except:
            install('PyMySQL')
        try:
            import sqlalchemy as sa
        except:
            install('SQLAlchemy')
            import sqlalchemy as sa
        if user is not None and password is None:
            url = f"mysql+pymysql://{user}@{host}:{port}"
        elif user is None and password is None:
            url = f"mysql+pymysql://{host}:{port}"
        elif user is not None and password is not None:
            url = f"mysql+pymysql://{user}:{password}@{host}:{port}"
        else:
            raise ValueError('mysql database address error.')
        self._params.engine = sa.create_engine(url, **kwargs)
        

class HiveIoDataset(DataSet):
    """Create a data set composed of hive database.

    Args:
        host: hive database host.
        port: hive database port.
        user: hive database user.
        password: hive database password.
    """
    def __init__(self, host, port=10000, user=None, password=None, **kwargs):
        super(HiveIoDataset, self).__init__()
        try:
            import pyhive
        except:
            install('pyhive')
        try:
            import sqlalchemy as sa
        except:
            install('SQLAlchemy')
            import sqlalchemy as sa
        if user is not None and password is None:
            url = f"hive://{user}@{host}:{port}"
        elif user is None and password is None:
            url = f"hive://{host}:{port}"
        elif user is not None and password is not None:
            url = f"hive://{user}:{password}@{host}:{port}"
        else:
            raise ValueError('hive database address error.')
        self._params.engine = sa.create_engine(url, **kwargs)
        
        


from linora.image._image_crop import *

class ImageCropAug(object):
    def __init__(self, image=None):
        self.image = image
        
    def crop(self, box, p=None):
        """Returns a rectangular region from this image. 
        
        The box is a 4-tuple defining the left, upper, right, and lower pixel coordinate.
        Args:
            box: The crop rectangle, as a (left, upper, right, lower)-tuple.
            p: probability that the image does this. Default value is 1.
        """
        if self._max_aug_nums>0:
            if self._nums>self._max_aug_nums:
                return self
            self._nums += 1
        if p is None:
            p = self._p
        self.image = crop(self.image, box, p)
        return self

    def crop_central(self, central_rate, p=None):
        """Crop the central region of the image.
    
        Remove the outer parts of an image but retain the central region of the image
        along each dimension. If we specify central_fraction = 0.5, this function
        returns the region marked with "X" in the below diagram.
           --------
          |        |
          |  XXXX  |
          |  XXXX  |
          |        |   where "X" is the central 50% of the image.
           --------

        This function works on either a single image (`image` is a 3-D Tensor), or a
        batch of images (`image` is a 4-D Tensor).

        Args:
            central_rate: if int float, should be in the interval (0, 1], fraction of size to crop.
                          if tuple list, randomly picked in the interval
                          `[central_rate[0], central_rate[1])`, value is fraction of size to crop.
            p: probability that the image does this. Default value is 1.
        """
        if self._max_aug_nums>0:
            if self._nums>self._max_aug_nums:
                return self
            self._nums += 1
        if p is None:
            p = self._p
        self.image = crop_central(self.image, central_rate, p)
        return self
    
    def crop_point(self, height_rate, width_rate, p=None):
        """Crop the any region of the image(s) and resize specify shape.
    
        Crop region area = height_rate * width_rate *image_height * image_width

        This function works on either a single image (`image` is a 3-D Tensor), or a
        batch of images (`image` is a 4-D Tensor).

        Args:
            height_rate: float, in the interval (0, 1].
            width_rate: float, in the interval (0, 1].
            p: probability that the image does this. Default value is 1.
        Raises:
            ValueError: if central_crop_fraction is not within (0, 1].
        """
        if self._max_aug_nums>0:
            if self._nums>self._max_aug_nums:
                return self
            self._nums += 1
        if p is None:
            p = self._p
        self.image = crop_point(self.image, height_rate, width_rate, p)
        return self



from linora import audio
from linora import chart
from linora import data
from linora import feature_column
from linora import feature_selection
from linora import gfile
from linora import image
from linora import metrics
from linora import parallel
from linora import param_search
from linora import sample
from linora import server
from linora import text
from linora import train
from linora import utils
from linora import vedio

__version__ = '2.0.0rc1'
__author__ = 'JinQing Lee'

def _version():
    print(f"""
------------------------------------------------------------------------------------
      Linora
--------------------
      Version      : --  {__version__}  --
      Author       : --  {__author__}  --
      License      : Apache-2.0
      Homepage     : https://github.com/Hourout/linora
      Description  : Simple and efficient tools for data mining and data analysis.
------------------------------------------------------------------------------------""")



import math
import collections

import numpy as np

from linora.text._word_processing import ngrams_iterator

__all__ = ['bleu_score']


def bleu_score(candidate_corpus, references_corpus, max_n=4, weights=[0.25] * 4):
    """Computes the BLEU score between a candidate translation corpus and a references
    translation corpus. Based on https://www.aclweb.org/anthology/P02-1040.pdf

    Args:
        candidate_corpus: an iterable of candidate translations. Each translation is an
            iterable of tokens
        references_corpus: an iterable of iterables of reference translations. Each
            translation is an iterable of tokens
        max_n: the maximum n-gram we want to use. E.g. if max_n=3, we will use unigrams,
            bigrams and trigrams
        weights: a list of weights used for each n-gram category (uniform by default)
    """

    assert max_n == len(weights), 'Length of the "weights" list has be equal to max_n'
    assert len(candidate_corpus) == len(
        references_corpus
    ), "The length of candidate and reference corpus should be the same"

    clipped_counts = np.zeros(max_n)
    total_counts = np.zeros(max_n)
    weights = np.array(weights)

    candidate_len = 0.0
    refs_len = 0.0

    for (candidate, refs) in zip(candidate_corpus, references_corpus):
        candidate_len += len(candidate)

        # Get the length of the reference that's closest in length to the candidate
        refs_len_list = [float(len(ref)) for ref in refs]
        refs_len += min(refs_len_list, key=lambda x: abs(len(candidate) - x))

        reference_counters = collections.Counter(tuple(x.split(" ")) for x in ngrams_iterator(refs[0], max_n))
        for ref in refs[1:]:
            reference_counters = reference_counters | collections.Counter(tuple(x.split(" ")) for x in ngrams_iterator(ref, max_n))

        candidate_counter = collections.Counter(tuple(x.split(" ")) for x in ngrams_iterator(candidate, max_n))

        clipped_counter = candidate_counter & reference_counters

        for ngram in clipped_counter:
            clipped_counts[len(ngram) - 1] += clipped_counter[ngram]

        for ngram in candidate_counter:
            total_counts[len(ngram) - 1] += candidate_counter[ngram]

    if min(clipped_counts) == 0:
        return 0.0
    else:
        pn = clipped_counts / total_counts
        log_pn = weights * np.log(pn)
        score = np.exp(sum(log_pn))

        bp = math.exp(min(1 - refs_len / candidate_len, 0))

        return bp * score.item()


from itertools import chain
from collections import Counter

import numpy as np

__all__ = ['select_best_length', 'word_to_index', 'sequence_word_index', 'index_vector_matrix',
           'sequence_index_word']


def select_best_length(sequence, sample_rate=0.8):
    """Select best length for sequence with keep rate.
    
    Args:
        sequence: pd.Series or np.array or List of lists, sample sequence.
        sample_rate: float, keep sample rate.
    Returns:
        int length, sample max keep length.
    """
    t = sorted(map(lambda x:len(x), sequence))
    return t[int(np.ceil(len(sequence)*sample_rate))-1]


def word_to_index(sequence):
    """Sequence word transfer to index.
    
    Args:
        sequence: pd.Series or np.array or List of lists, sample sequence.
    Returns:
        dict, {'positive':{word: index}，'negative':{index:word}}, The higher the frequency, the higher the ranking.
    """
    t = Counter(chain.from_iterable(sequence))
    t = sorted(t, key=t.get, reverse=True)
    return {'positive':{v: k + 1 for k, v in enumerate(t)}, 'negative':{k + 1:v for k, v in enumerate(t)}}


def sequence_word_index(sequence, word_index_dict, pad_value=0):
    """Sequence word transfer to sequence index.
    
    Args:
        sequence: pd.Series or np.array or List of lists, sample sequence.
        word_index_dict: dict, {word: index}.
        pad_value: fillna value, if word not in word_index_dict, fillna it.
    Returns:
        List of lists, sequence word transfer to sequence index list.
    """
    if isinstance(sequence[0], str):
        return [word_index_dict.get(j, pad_value) for j in sequence]
    return [[word_index_dict.get(j, pad_value) for j in i] for i in sequence]


def sequence_index_word(sequence, index_word_dict, pad_value=' ', join=None):
    """Sequence index transfer to sequence word.
    
    Args:
        sequence: pd.Series or np.array or List of lists, sample sequence.
        index_word_dict: dict, {index: word}.
        pad_value: fillna value, if word not in index_word_dict, fillna it.
        join: str, whether to merge the converted words. represents the merge interval
    Returns:
        List of lists, sequence index transfer to sequence word list.
    """
    if isinstance(sequence[0], int):
        if join=='':
            return ''.join([index_word_dict.get(j, pad_value) for j in sequence])
        if join:
            return f'{join}'.join([index_word_dict.get(j, pad_value) for j in sequence])
        return [index_word_dict.get(j, pad_value) for j in sequence]
    if join=='':
            return [''.join([index_word_dict.get(j, pad_value) for j in i]) for i in sequence]
    if join:
        return [f'{join}'.join([index_word_dict.get(j, pad_value) for j in i]) for i in sequence]
    return [[index_word_dict.get(j, pad_value) for j in i] for i in sequence]


def index_vector_matrix(word_index_dict, word_vector_dict, embed_dim=300, initialize='norm', dtype='float32'):
    """Make index vector matrix with shape `(len(word_index_dict), embed_dim)`.
    
    Args:
        word_index_dict: dict, {word: index}.
        word_vector_dict: dict, {word: vector}.
        embed_dim: embedding dim.
        initialize: initialize method, 'zero' or 'one' or 'norm'.
        dtype: default 'float32', data types.
    Returns:
        a index vector matrix with shape `(len(word_index_dict), embed_dim)`.
    """
    word_count = len(word_index_dict)+1
    init = {'zero':lambda x,y:np.zeros((x,y), dtype=dtype),
            'one':lambda x,y:np.ones((x,y), dtype=dtype),
            'norm':lambda x,y:np.random.normal(size=(x,y), dtype=dtype)}
    index_embed_matrix = init[initialize](word_count, embed_dim)
    for word, index in word_index_dict.items():
        l = len(word_vector_dict[word])
        if l>embed_dim:
            index_embed_matrix[index, :] = word_vector_dict[word][:embed_dim]
        else:
            index_embed_matrix[index, :l] = word_vector_dict[word]
    return index_embed_matrix



import numpy as np


class Hist2d():
    def add_hist2d(self, name, xdata, ydata, **kwargs):
        """A scatter plot of *y* vs. *x* with varying marker size and/or color.
        
        Args:
            name: data name.
            xdata: x-axis data.
            ydata: y-axis data.
        """
        if 'color' not in kwargs:
            kwargs['color'] = self._params.color.pop(0)[1]#tuple([round(np.random.uniform(0, 1),1) for _ in range(3)])
        else:
            if isinstance(kwargs['color'], dict):
                kwargs['color'] = kwargs.pop('color')['mode']
        self._params.ydata[name]['kwargs'] = kwargs
        self._params.ydata[name]['xdata'] = xdata
        self._params.ydata[name]['ydata'] = ydata
        self._params.ydata[name]['plotmode'] = 'hist2d'
        self._params.ydata[name]['plotfunc'] = self._execute_plot_hist2d
        return self
    
    def _execute_plot_hist2d(self, fig, ax, i, j):
        ax_plot = ax.hist2d(j['xdata'], j['ydata'], **j['kwargs'])



import itertools

import numpy as np
from PIL import ImageDraw, ImageChops

from linora.image._image_util import array_to_image
from linora.image._image_rgb import _fill_color

__all__ = ['draw_box', 'draw_point', 'mask', 'mask_box', 'draw_line', 'draw_keypoints', 'draw_segmentation_masks', 'stripe']


def draw_point(image, points, size=0, color=None):
    """Draws a point.
    
    Args:
        image: a PIL instance.
        points: Sequence of either 2-tuples like [(x, y), (x, y), ...] or numeric values like [x, y, x, y, ...].
        size: point size.
        color: str or tuple or la.image.RGBMode, rgb color, point color.
    returns: 
        a PIL instance.
    """
    image2 = image.copy()
    draw = ImageDraw.Draw(image2)
    point = list(itertools.chain.from_iterable(points)) if isinstance(points[0], (list, tuple)) else points
    for i in range(int(len(point)/2)):
        axis = list(itertools.product(range(int(point[i*2]-size), int(point[i*2]+size+1)), 
                                      range(int(point[i*2+1]-size), int(point[i*2+1]+size+1))))
        color1 = _fill_color(image, color)
        draw.point(axis, color1)
    return image2


def mask(image, size, max_num, random=True, color=None, p=1):
    """Draws a mask.
    
    Args:
        image: a PIL instance.
        size: list or tuple, mask size, [height, width]. if int, transform [size, size].
        max_num: int, max mask number.
                 if tuple or list, randomly picked in the interval `[max_num[0], max_num[1])`.
        random: bool, whether the mask position is random.
        color: str or tuple or la.image.RGBMode, rgb color, mask fill color.
        p: probability that the image does this. Default value is 1.
    returns: 
        a PIL instance.
    """
    if np.random.uniform()>p:
        return image
    if isinstance(max_num, (list, tuple)):
        max_num = int(np.random.uniform(max_num[0], max_num[1]))
    if isinstance(size, (int, float)):
        size = [int(size), int(size)]
    image2 = image.copy()
    draw = ImageDraw.Draw(image2)
    if random:
        for i in range(max_num):
            axis = (np.random.randint(0, image.width-size[1]), np.random.randint(0, image.height-size[0]))
            axis = [axis, (axis[0]+size[1], axis[1]+size[0])]
            color1 = _fill_color(image, color)
            draw.rectangle(axis, fill=color1, width=0)
    else:
        width_num = min(int(image.width/size[1]*0.6), int(max_num**0.5))
        height_num = min(int(max_num/width_num), int(image.height/size[0]*0.6))
        width_pix = int((image.width-width_num*size[1])/(width_num+1))
        height_pix = int((image.height-height_num*size[0])/(height_num+1))
        for i in range(width_num):
            for j in range(height_num):
                axis = [width_pix*(i+1)+size[1]*i, height_pix*(j+1)+size[0]*j, 
                        width_pix*(i+1)+size[1]*(i+1), height_pix*(j+1)+size[0]*(j+1)]
                color1 = _fill_color(image, color)
                draw.rectangle(axis, fill=color1, width=0)
    return image2


def mask_box(image, boxes, fill='inner', color=None, p=1):
    """Fill color inside or outside a rectangular area.
    
    Args:
        image: a PIL  instance.
        boxes: box axis with same [[x1,y1,x2,y2],...]
        fill: 'inner' or 'outer', color fill position.
        color: str or tuple or la.image.RGBMode, rgb color, box line color.
        p: probability that the image does this. Default value is 1.
    Returns:
        a PIL  instance.
    """
    if np.random.uniform()>p:
        return image
    if isinstance(boxes[0], int):
        boxes = [boxes]
    color = _fill_color(image, color)
    if fill=='inner':
        image1 = image.copy()
        for i in boxes:
            image1.paste(Image.new(image.mode, (i[2]-i[0], i[3]-i[1]), color=color), (i[0], i[1]))
    else:
        image1 = Image.new(image.mode, image.size, color=color)
        for i in boxes:
            image1.paste(image.crop(i), (i[0], i[1]))
    return image1


def stripe(image, width=4, mode=1, color=None, p=1):
    """vertically and horizontally line apply to image.
    
    Args:
        image: a numpy array.
        width: int, line spacing width.
        mode: 0 is vertically and 1 is horizontally.
        color: int or list, line color.
        p: probability that the image does this. Default value is 1.
    returns: 
        a numpy array.
    """
    if np.random.uniform()>p:
        return image
    image1 = image.copy()
    color = _fill_color(image1, color)
    if mode:
        image1[::width,:] = color
    else:
        image1[:,::width] = color
    return image1


def draw_box(image, boxs, fill_color=None, line_color=None, width=1):
    """Draws a polygon.
    
    The polygon outline consists of straight lines between the given coordinates, 
    plus a straight line between the last and the first coordinate.
    
    Args:
        image: a PIL instance.
        boxs: If it is a single sample, like [(x1, y1), (x2, y2), ...] or like [x1, y1, x2, y2, ...].
              If there are multiple samples, like [[(x1, y1), (x2, y2), ...]] or like [[x1, y1, x2, y2, ...]].
              special, like [(x1, y1), (x2, y2)] or like [x1, y1, x2, y2], draw a rectangle from the top left to the bottom right.
              special, like [x,y], or like [[x,y]], Translate into [x, x, y, y], draw a rectangle from the top left to the bottom right.
        fill_color: str or tuple or la.image.RGBMode, rgb color, box fill color.
        line_color: str or tuple or la.image.RGBMode, rgb color, box line color.
        width: box line width.
    returns: 
        a PIL instance.
    """
    image2 = image.copy()
    draw = ImageDraw.Draw(image2)
    if isinstance(boxs[0], (int, float)):
        if len(boxs)==2:
            boxs = [[min(boxs), min(boxs), min(boxs), max(boxs), max(boxs), max(boxs), max(boxs), min(boxs)]]
        elif len(boxs)==4:
            boxs = [[min(boxs[0], boxs[2]), min(boxs[1], boxs[3]), max(boxs[0], boxs[2]), min(boxs[1], boxs[3]),
                     max(boxs[0], boxs[2]), max(boxs[1], boxs[3]), min(boxs[0], boxs[2]), max(boxs[1], boxs[3])]]
        elif len(boxs)%2 == 0:
            boxs = [boxs]
        else:
            raise ValueError('boxs axis error')
    elif isinstance(boxs[0], (list, tuple)):
        if len(boxs)==1 and len(boxs[0])==2 and isinstance(boxs[0][0], (int, float)):
            boxs = [[min(boxs[0]), min(boxs[0]), min(boxs[0]), max(boxs[0]), max(boxs[0]), max(boxs[0]), max(boxs[0]), min(boxs[0])]]
        elif len(boxs)==2 and len(boxs[0])==2:
            boxs = [boxs[0][0], boxs[0][1], boxs[1][0], boxs[1][1]]
            boxs = [[min(boxs[0], boxs[2]), min(boxs[1], boxs[3]), max(boxs[0], boxs[2]), min(boxs[1], boxs[3]),
                     max(boxs[0], boxs[2]), max(boxs[1], boxs[3]), min(boxs[0], boxs[2]), max(boxs[1], boxs[3])]]
        elif len(boxs)>2 and len(boxs[0])==2:
            boxs = [[j for i in boxs for j in i]]
        else:
            boxs = [list(itertools.chain.from_iterable(i)) if isinstance(i[0], (list, tuple)) else i for i in boxs]
            for i in boxs:
                if len(i)%2 != 0 or len(i)<6:
                    raise ValueError('boxs axis error')
    else:
            raise ValueError('boxs axis error')
    for i in boxs:
        if isinstance(fill_color, dict):
            fill_color = fill_color['mode']
        line_color1 = _fill_color(image, line_color)
        draw.polygon(i, fill=fill_color, outline=line_color1, width=width)
    return image2


def draw_line(image, axis, width=1, color=None):
    """Draws a line.
    
    Args:
        image: a PIL instance.
        axis: Sequence of either 2-tuples like [(x, y), (x, y), ...] or numeric values like [x, y, x, y, ...].
        width: int, the line width, in pixels.
        color: str or tuple or la.image.RGBMode, rgb color, line color.
    returns: 
        a PIL instance.
    """
    image2 = image.copy()
    draw = ImageDraw.Draw(image2)
    axis = list(itertools.chain.from_iterable(axis)) if isinstance(axis[0], (list, tuple)) else axis
    for i in range(int(len(axis)/2-1)):
        line = [axis[i*2], axis[i*2+1], axis[i*2+2], axis[i*2+3]]
        color1 = _fill_color(image, color)
        draw.line(line, fill=color1, width=width)
    return image2


def draw_keypoints(image, keypoints, connectivity=None, point_width=3, line_width=2, point_color=None, line_color=None):
    """Draws Keypoints on given RGB image.
    
    Args:
        image: a PIL instance.
        keypoints: list of shape (num_instances, K, 2) the K keypoints location for each of the N instances,
                   in the format [x, y].
        connectivity: A List of tuple where, each tuple contains pair of keypoints to be connected.
        point_width: Integer denoting radius of keypoint.
        line_width: Integer denoting width of line connecting keypoints.
        point_color: str or tuple or la.image.RGBMode, rgb color, key points color.
        line_color: str or tuple or la.image.RGBMode, rgb color, line color.
    Returns:
        a PIL instance.
    """    
    image2 = image.copy()
    draw = ImageDraw.Draw(image2)
    if isinstance(keypoints[0][0], (int, float)):
        keypoints = [keypoints]
    for kpt_id, kpt_inst in enumerate(keypoints):
        for inst_id, kpt in enumerate(kpt_inst):
            x1 = kpt[0] - point_width
            x2 = kpt[0] + point_width
            y1 = kpt[1] - point_width
            y2 = kpt[1] + point_width
            point_color = _fill_color(image, point_color)
            draw.ellipse([x1, y1, x2, y2], fill=point_color, outline=None, width=0)
        if connectivity:
            for connection in connectivity:
                start_pt_x = kpt_inst[connection[0]][0]
                start_pt_y = kpt_inst[connection[0]][1]
                end_pt_x = kpt_inst[connection[1]][0]
                end_pt_y = kpt_inst[connection[1]][1]
                color = _fill_color(image, line_color)
                draw.line(((start_pt_x, start_pt_y), (end_pt_x, end_pt_y)), fill=color, width=line_width)
    return image2


def draw_segmentation_masks(mask, image=None, alpha=0.8, color=None):
    """Draws segmentation masks on given RGB image.
    
    Args:
        mask: a numpy array, shape is (H, W).
        image: a PIL instance.
        alpha: float number between 0 and 1 denoting the transparency of the masks.
               0 means full transparency, 1 means no transparency.
        color: str or tuple or la.image.RGBMode, rgb color, point color.
               List containing the colors of the masks.
    returns: 
        a PIL instance.
    """
    canvas = np.zeros((mask.shape[0], mask.shape[1], 3))
    if color is not None:
        assert len(color)==mask.max()+1, 'color size not equal mask nunique size.'
    else:
        color = tuple([tuple([np.random.randint(0, 256) for j in range(3)]) for i in range(int(mask.max()+1))])
    for label in range(int(mask.max()+1)):
        if canvas[mask == label].size:
            canvas[mask == label] = color[label]
    canvas = array_to_image(canvas)
    if image is None:
        return canvas
    if image.size!=canvas.size:
        return ImageChops.blend(image.resize(canvas.size), canvas, alpha=alpha)
    return ImageChops.blend(image, canvas, alpha=alpha)


import random

import numpy as np
from PIL import Image, ImageStat

from linora.image._image_draw import draw_point

__all__ = ['mosaic', 'noise_saltpepper', 'noise_gaussian', 'noise_laplace', 'noise_poisson', 
           'noise_uniform', 'noise_speckle', 'noise_impulse'
          ]


def mosaic(image, size=(80,80), block=0, axis=None, prob=0.3, p=1):
    """mosaic noise apply to image.
    
    Args:
        image: a PIL instance.
        size: if int or float, xsize=ysize, how many mosaic blocks the image is divided into.
              if 2-tuple, (xsize, ysize), 
              if 4-tuple, xsize in (size[0], size[1]) and  ysize in (size[2], size[3]).
        block: mosaic area block
        axis: list or tuple, one or more mosaic center axis point.
        prob: probability of numbers of mosaci.
        p: probability that the image does this. Default value is 1.
    Return:
        a PIL instance.
    """
    if isinstance(size, (int, float)):
        size = (size, size)
    elif isinstance(size, (list, tuple)):
        if len(size)==4:
            size = (np.random.uniform(size[0], size[1]), np.random.uniform(size[2], size[3]))
        elif len(size)!=2:
            raise ValueError('`size` value format error.')
    else:
        raise ValueError('`size` value format error.')
    size = (max(2, int(size[0])), max(2, int(size[1])))
    
    if axis is not None:
        if isinstance(axis[0], (list, tuple)):
            block = max(block, len(axis))
            axis_num = len(axis)
        else:
            block = max(block, 1)
            axis_num = 1
            axis = [axis]

    w = image.size[0]//size[0]
    h = image.size[1]//size[1]
    image2 = image.copy()
    if block==0:
        axis_list = [[i*w, j*h, i*w+w, j*h+h] for i in range(size[0]) for j in range(size[1])]
        s = np.random.choice(range(size[0]*size[1]), max(int(size[0]*size[1]*prob), 2), replace=False)
        for i in s:
            image2.paste(Image.new(image.mode, (w,h), tuple([int(j) for j in ImageStat.Stat(image.crop(axis_list[i])).mean])), axis_list[i])
    else:
        num = max(int(size[0]*size[1]*prob/block), 2)
        edge = np.sqrt(num)
        
        for b in range(int(block)):
            if axis is not None:
                if b<=axis_num-1:
                    axis_list = [axis[b][0]/w, axis[b][1]/h]
                else:
                    axis_list = [np.random.uniform(edge/2/size[0], 1-edge/2/size[0])*size[0], 
                                 np.random.uniform(edge/2/size[1], 1-edge/2/size[0])*size[1]]
            else:
                axis_list = [np.random.uniform(edge/2/size[0], 1-edge/2/size[0])*size[0], 
                             np.random.uniform(edge/2/size[1], 1-edge/2/size[0])*size[1]]
            axis_prob = []
            axis_must = []
            for i in range(int(axis_list[0]-edge/2), int(axis_list[0]+edge/2)+1):
                for j in range(int(axis_list[1]-edge/2), int(axis_list[1]+edge/2)+1):
                    if i<int(axis_list[0]-edge/3) or i>int(axis_list[0]+edge/3) or j<int(axis_list[1]-edge/3) or j>int(axis_list[1]+edge/3):
                        axis_prob.append([i*w, j*h, i*w+w, j*h+h])
                    else:
                        axis_must.append([i*w, j*h, i*w+w, j*h+h])
            s = np.random.choice(range(len(axis_prob)), int(len(axis_prob)*0.8), replace=False)
            for i in s:
                image2.paste(Image.new(image.mode, (w,h), tuple([int(j) for j in ImageStat.Stat(image.crop(axis_prob[i])).mean])), axis_prob[i])
            for i in axis_must:
                image2.paste(Image.new(image.mode, (w,h), tuple([int(j) for j in ImageStat.Stat(image.crop(i)).mean])), i)
    return image2


def noise_saltpepper(image, white_prob=0.025, black_prob=0.025, p=1):
    """Mask salt pepper noise apply to image.
    
    The salt-pepper noise is based on the signal-to-noise ratio of the image,
    randomly generating the pixel positions in some images all channel,
    and randomly assigning these pixels to 0 or 255.
    
    Args:
        image: a PIL instance.
        white_prob: white pixel prob.
        black_prob: black pixel prob.
        p: probability that the image does this. Default value is 1.
    Returns:
        a PIL instance.
    """
    if np.random.uniform()>p:
        return image
    img = image.copy()
    axis = [(i,j) for i in range(image.width) for j in range(image.height)]
    np.random.shuffle(axis)
    if isinstance(white_prob, (tuple, list)):
        white_prob = np.random.uniform(white_prob[0], white_prob[1])
    if isinstance(black_prob, (tuple, list)):
        black_prob = np.random.uniform(black_prob[0], black_prob[1])
    if white_prob>0:
        axis_white = axis[0:int(len(axis)*white_prob)]
        pixel = (255,)*len(image.getbands())
        for i in axis_white:
            img.putpixel(i, pixel)
    if black_prob>0:
        axis_black = axis[-int(len(axis)*(black_prob)):]
        pixel = (0,)*len(image.getbands())
        for i in axis_black:
            img.putpixel(i, pixel)
    return img


def _noise(image, function, wise='pixel', scale=1, prob=0.2, p=1, **kwargs):
    if np.random.uniform()>p:
        return image
    if wise=='pixel':
        return image.point(lambda x:function(x, **kwargs)*scale if np.random.uniform()<prob else x)
    elif wise=='channel':
        split = list(image.split())
        for i in range(len(split)):
            if np.random.uniform()<prob:
                split[i] = split[i].point(lambda x:function(x, **kwargs)*scale)
        return Image.merge(image.mode, split)
    elif wise=='pixelchannel':
        img = image.point(lambda x:function(x, **kwargs)*scale)
        axis = [(i,j) for i in range(image.width) for j in range(image.height)]
        np.random.shuffle(axis)
        if prob>0.5:
            axis = axis[:int(len(axis)*(1-prob))]
            for i in axis:
                img.putpixel(i, image.getpixel(i))
            return img
        else:
            axis = axis[:int(len(axis)*prob)]
            for i in axis:
                image.putpixel(i, img.getpixel(i))
            return image
    elif isinstance(wise, (list, tuple)):
        split = list(image.split())
        for i in wise:
            if np.random.uniform()<prob:
                split[i] = split[i].point(lambda x:function(x, **kwargs)*scale)
        return Image.merge(image.mode, split)
    else:
        raise ValueError("`wise` should be 'pixel' or 'channel' or 'pixelchannel' or list of channel.")


def noise_gaussian(image, mean=0, std=30, wise='pixel', scale=1, prob=0.2, p=1):
    """gaussian distribution noise apply to image.
    
    Args:
        image: A PIL instance.
        mean: if int or float, value is gaussian distribution mean.
              if tuple or list, randomly picked in the interval `[mean[0], mean[1])`.
        std: if int or float, value is gaussian distribution std.
             if tuple or list, randomly picked in the interval `[std[0], std[1])`.
        wise: 'pixel' or 'channel' or 'pixelchannel' or list of channel, method of applying noise.
        scale: if int or float, value multiply with noise.
               if tuple or list, randomly picked in the interval `[scale[0], scale[1])`.
        prob: probability of every pixel or channel being changed.
        p: probability that the image does this. Default value is 1.
    Returns:
        A PIL instance.
    """
    def func(x, mean, std):
        return np.random.normal(mean, std)
    if isinstance(mean, (tuple, list)):
        mean = np.random.uniform(mean[0], mean[1])
    if isinstance(std, (tuple, list)):
        std = np.random.uniform(std[0], std[1])
    if isinstance(scale, (tuple, list)):
        scale = np.random.uniform(scale[0], scale[1])
    return _noise(image, func, wise=wise, scale=scale, prob=prob, p=p, mean=mean, std=std)


def noise_laplace(image, mean=0, lam=30, wise='pixel', scale=1, prob=0.2, p=1):
    """laplace distribution noise apply to image.
    
    Args:
        image: A PIL instance.
        mean: if int or float, value is laplace distribution mean.
              if tuple or list, randomly picked in the interval `[mean[0], mean[1])`.
        lam: if int or float, value is laplace distribution lam.
             if tuple or list, randomly picked in the interval `[lam[0], lam[1])`.
        wise: 'pixel' or 'channel' or 'pixelchannel' or list of channel, method of applying noise.
        scale: if int or float, value multiply with noise.
               if tuple or list, randomly picked in the interval `[scale[0], scale[1])`.
        prob: probability of every pixel or channel being changed.
        p: probability that the image does this. Default value is 1.
    Returns:
        A PIL instance.
    """
    def func(x, mean, lam):
        return np.random.laplace(mean, lam)
    if isinstance(mean, (tuple, list)):
        mean = np.random.uniform(mean[0], mean[1])
    if isinstance(lam, (tuple, list)):
        lam = np.random.uniform(lam[0], lam[1])
    if isinstance(scale, (tuple, list)):
        scale = np.random.uniform(scale[0], scale[1])
    return _noise(image, func, wise=wise, scale=scale, prob=prob, p=p, mean=mean, lam=lam)


def noise_poisson(image, lam=30, wise='pixel', scale=1, prob=0.2, p=1):
    """poisson distribution noise apply to image.
    
    Args:
        image: A PIL instance.
        lam: if int or float, value is poisson distribution lam.
             if tuple or list, randomly picked in the interval `[lam[0], lam[1])`.
        wise: 'pixel' or 'channel' or 'pixelchannel' or list of channel, method of applying noise.
        scale: if int or float, value multiply with noise.
               if tuple or list, randomly picked in the interval `[scale[0], scale[1])`.
        prob: probability of every pixel or channel being changed.
        p: probability that the image does this. Default value is 1.
    Returns:
        A PIL instance.
    """
    def func(x, lam):
        return np.random.poisson(lam)
    if isinstance(lam, (tuple, list)):
        lam = np.random.uniform(lam[0], lam[1])
    if isinstance(scale, (tuple, list)):
        scale = np.random.uniform(scale[0], scale[1])
    return _noise(image, func, wise=wise, scale=scale, prob=prob, p=p, lam=lam)


def noise_uniform(image, lower=-50, upper=50, wise='pixel', scale=1, prob=0.2, p=1):
    """uniform distribution noise apply to image.
    
    Args:
        image: A PIL instance.
        lower: if int or float, value is uniform distribution lower.
               if tuple or list, randomly picked in the interval `[lower[0], lower[1])`.
        upper: if int or float, value is uniform distribution upper.
               if tuple or list, randomly picked in the interval `[upper[0], upper[1])`.
        wise: 'pixel' or 'channel' or 'pixelchannel' or list of channel, method of applying noise.
        scale: if int or float, value multiply with noise.
               if tuple or list, randomly picked in the interval `[scale[0], scale[1])`.
        prob: probability of every pixel or channel being changed.
        p: probability that the image does this. Default value is 1.
    Returns:
        A PIL instance.
    """
    def func(x, lower, upper):
        return np.random.uniform(lower, upper)
    if isinstance(lower, (tuple, list)):
        lower = np.random.uniform(lower[0], lower[1])
    if isinstance(upper, (tuple, list)):
        upper = np.random.uniform(upper[0], upper[1])
    if isinstance(scale, (tuple, list)):
        scale = np.random.uniform(scale[0], scale[1])
    return _noise(image, func, wise=wise, scale=scale, prob=prob, p=p, lower=lower, upper=upper)


def noise_speckle(image, wise='pixel', prob=0.2, p=1):
    """speckle noise apply to image.
    
    pixel = pixel+pixel*gaussian
    
    Args:
        image: A PIL instance.
        prob: probability of every pixel or channel being changed.
        p: probability that the image does this. Default value is 1.
    Returns:
        A PIL instance.
    """
    def func(x):
        return x*(1+np.random.normal())
    return _noise(image, func, wise=wise, scale=1, prob=prob, p=p)


def noise_impulse(image, binomial=0.5, wise='pixel', prob=0.2, p=1):
    """impulse noise apply to image.
        
    Args:
        image: A PIL instance.
        binomial: if int or float, value is binomial distribution prob.
                  if tuple or list, randomly picked in the interval `[binomial[0], binomial[1])`.
        wise: 'pixel' or 'channel' or 'pixelchannel' or list of channel, method of applying noise.
        prob: probability of every pixel or channel being changed.
        p: probability that the image does this. Default value is 1.
    Returns:
        A PIL instance.
    """
    def func(x, binomial):
        mask = np.random.binomial(n=1, p=binomial)
        return x*(1 - mask) + np.random.randint(256) * mask
    if isinstance(binomial, (tuple, list)):
        binomial = np.random.uniform(binomial[0], binomial[1])
    return _noise(image, func, wise=wise, scale=1, prob=prob, p=p, binomial=binomial)







import numpy as np
import pandas as pd

__all__ = ['divergence_kl', 'divergence_js', 'divergence_F',
           'mutual_information', 'mutual_information_rate', 
           'pointwise_mutual_information', 'pointwise_mutual_information_rate'
          ]


def divergence_kl(x, y, continuous=False, bucket_num=1000):
    """kl divergence.
    
    Args:
        x: pd.Series or array or list, sample n dim feature value.
        y: pd.Series or array or list, sample n dim feature value.
        continuous: bool, default False, whether it is a continuous value. if True, normalized x and y.
        bucket_num: int, default 1000, if continuous is True, perform bucket operations on features.
    Returns:
        kl divergence value.
    """
    x = pd.Series(x)
    y = pd.Series(y)
    if continuous:
        x = (x-min(x.min(), y.min()))/(max(x.max(), y.max())-min(x.min(), y.min()))
        y = (y-min(x.min(), y.min()))/(max(x.max(), y.max())-min(x.min(), y.min()))
        x = pd.cut(x, bucket_num, labels=range(bucket_num))
        y = pd.cut(y, bucket_num, labels=range(bucket_num))
    t = x.value_counts(normalize=True).reset_index().merge(y.value_counts(normalize=True).reset_index(), on='index', how='left').fillna(0.)
    t.columns = ['label', 'prob_x', 'prob_y']
    t = np.sum(np.log((t.prob_x+0.00001)/(t.prob_y+0.00001))*t.prob_x)
    return t


def divergence_js(x, y, continuous=False, bucket_num=1000):
    """js divergence.
    
    Args:
        x: pd.Series or array or list, sample n dim feature value.
        y: pd.Series or array or list, sample n dim feature value.
        continuous: default False, whether it is a continuous value. if True, normalized x and y.
        bucket_num: default 1000, if continuous is True, perform bucket operations on features.
    Returns:
        js divergence value.
    """
    x = pd.Series(x)
    y = pd.Series(y)
    if continuous:
        x = (x-min(x.min(), y.min()))/(max(x.max(), y.max())-min(x.min(), y.min()))
        y = (y-min(x.min(), y.min()))/(max(x.max(), y.max())-min(x.min(), y.min()))
        x = pd.cut(x, bucket_num, labels=range(bucket_num))
        y = pd.cut(y, bucket_num, labels=range(bucket_num))
    t = x.value_counts(normalize=True).reset_index().merge(y.value_counts(normalize=True).reset_index(), on='index', how='outer').fillna(0.)
    t.columns = ['label', 'prob_x', 'prob_y']
    t['prob_m'] = (t.prob_x+t.prob_y)/2
    t1 = np.sum(np.log((t.prob_x+0.00001)/(t.prob_m+0.00001))*t.prob_x)
    t2 = np.sum(np.log((t.prob_y+0.00001)/(t.prob_m+0.00001))*t.prob_y)
    return 0.5*(t1+t2)


def divergence_F(x, y):
    """F divergence.
    
    Args:
        x: pd.Series or array or list, sample n dim feature value.
        y: pd.Series or array or list, sample n dim feature value.
    Returns:
        F divergence value.
    """
    x = np.array(x)
    y = np.array(y)
    value = y*x
    x[np.where(x==0)] = 1
    y[np.where(y==0)] = 1
    return np.sum(value/y*np.log(x/y))


def mutual_information(feature1, feature2):
    """mutual information.
    
    Args:
        feature1: pd.Series, sample feature value.
        feature2: pd.Series, sample feature value.
    Returns:
        mutual information value.
    """
    t = feature1.value_counts(normalize=True).reset_index()
    t.columns = ['label', 'label_mean']
    t['temp'] = 0
    t1 = feature2.value_counts(normalize=True).reset_index()
    t1.columns = ['label1', 'label1_mean']
    t1['temp'] = 0
    t = t.merge(t1, on='temp')
    t['label2'] = t.label.astype(str)+'-'+t.label1.astype(str)
    t2 = (feature1.astype(str)+'-'+feature2.astype(str)).value_counts(normalize=True).reset_index()
    t2.columns = ['label2', 'label2_mean']
    t = t.merge(t2, on='label2', how='left').dropna()
    t['mutual_information'] = t.label2_mean*np.log2(t.label2_mean/t.label_mean/t.label1_mean)
    return t.mutual_information.sum()


def pointwise_mutual_information(feature1, feature2):
    """pointwise mutual information.
    
    Args:
        feature1: pd.Series, sample feature value.
        feature2: pd.Series, sample feature value.
    Returns:
        pointwise mutual information value.
    """
    t = feature1.value_counts(normalize=True).reset_index()
    t.columns = ['label', 'label_mean']
    t['temp'] = 0
    t1 = feature2.value_counts(normalize=True).reset_index()
    t1.columns = ['label1', 'label1_mean']
    t1['temp'] = 0
    t = t.merge(t1, on='temp')
    t['label2'] = t.label.astype(str)+'-'+t.label1.astype(str)
    t2 = (feature1.astype(str)+'-'+feature2.astype(str)).value_counts(normalize=True).reset_index()
    t2.columns = ['label2', 'label2_mean']
    t = t.merge(t2, on='label2', how='left').dropna()
    return np.power(np.cumprod(np.log2(t.label2_mean/t.label_mean/t.label1_mean)).max(), 1/len(t))


def mutual_information_rate(feature1, feature2):
    """mutual information rate.
    
    Args:
        feature1: pd.Series, sample feature value.
        feature2: pd.Series, sample feature value.
    Returns:
        mutual information rate value.
    """
    return mutual_information(feature1, feature2)/min(mutual_information(feature1, feature1), mutual_information(feature2, feature2))


def pointwise_mutual_information_rate(feature1, feature2):
    """pointwise mutual information rate.
    
    Args:
        feature1: pd.Series, sample feature value.
        feature2: pd.Series, sample feature value.
    Returns:
        pointwise mutual information rate value.
    """
    return pointwise_mutual_information(feature1, feature2)/min(pointwise_mutual_information(feature1, feature1), pointwise_mutual_information(feature2, feature2))



from linora.image._image_filter import *

class ImageFilterAug(object):
    def __init__(self, image=None):
        self.image = image
    
    def filters(self, mode=FilterMode.Mean, p=None, **kwarg):
        """
        la.image.FilterMode.Box:
            Blurs the image by setting each pixel to the average value of the pixels 
            in a square box extending radius pixels in each direction. 
            Supports float radius of arbitrary size. Uses an optimized implementation 
            which runs in linear time relative to the size of the image for any radius value.

            you should append param `radius`
            radius: Size of the box in one direction. Radius 0 does not blur, returns an identical image. 
                    Radius 1 takes 1 pixel in each direction, i.e. 9 pixels in total.
            eg.
            la.image.filters(image, mode=la.image.FilterMode.Box, radius=2)

        la.image.FilterMode.Gaussian:
            Gaussian blur filter.

            you should append param `radius`
            radius: Blur radius.
            eg.
            la.image.filters(image, mode=la.image.FilterMode.Gaussian, radius=2)

        la.image.FilterMode.Unsharpmask:
            Unsharp mask filter.
            See Wikipedia’s entry on digital unsharp masking for an explanation of the parameters.

            you should append param `radius`,`percent`,`threshold`
            radius: Blur radius.
            percent: Unsharp strength, in percent
            threshold: Threshold controls the minimum brightness change that will be sharpened
            eg.
            la.image.filters(image, mode=la.image.FilterMode.Unsharpmask, radius=2, percent=150, threshold=3)

        la.image.FilterMode.Rank:
            Create a rank filter. The rank filter sorts all pixels in a window of the given size, and returns the rank’th value.

            you should append param `size`,`rank`
            size: The kernel size, in pixels.
            rank: What pixel value to pick. Use 0 for a min filter, 
                  size * size / 2 for a median filter, size * size - 1 for a max filter, etc.
            eg.
            la.image.filters(image, mode=la.image.FilterMode.Rank, size, rank)

        la.image.FilterMode.Median:
            Create a median filter. Picks the median pixel value in a window with the given size.

            you should append param `size`
            size: The kernel size, in pixels.
            eg.
            la.image.filters(image, mode=la.image.FilterMode.Median, size=3)

        la.image.FilterMode.Min:
            Create a min filter. Picks the lowest pixel value in a window with the given size.

            you should append param `size`
            size: The kernel size, in pixels.
            eg.
            la.image.filters(image, mode=la.image.FilterMode.Min, size=3)

        la.image.FilterMode.Max:
            Create a max filter. Picks the largest pixel value in a window with the given size.

            you should append param `size`
            size: The kernel size, in pixels.
            eg.
            la.image.filters(image, mode=la.image.FilterMode.Max, size=3)

        la.image.FilterMode.Mode:
            Create a mode filter. Picks the most frequent pixel value in a box with the given size. 

            Pixel values that occur only once or twice are ignored; 
            if no pixel value occurs more than twice, the original pixel value is preserved.

            you should append param `size`
            size: The kernel size, in pixels.
            eg.
            la.image.filters(image, mode=la.image.FilterMode.Mode, size=3)

        la.image.FilterMode.Mean:
            Normal blur.
            eg.
            la.image.filters(image, mode=la.image.FilterMode.Mean)

        la.image.FilterMode.CONTOUR:
            contour blur.
            eg.
            la.image.filters(image, mode=la.image.FilterMode.CONTOUR)

        la.image.FilterMode.DETAIL:
            detail blur.
            eg.
            la.image.filters(image, mode=la.image.FilterMode.DETAIL)

        la.image.FilterMode.EDGE_ENHANCE:
            Edge enhancement blur.
            eg.
            la.image.filters(image, mode=la.image.FilterMode.EDGE_ENHANCE)
            
        la.image.FilterMode.EDGE_ENHANCE_MORE:
            Edge enhancement more blur.
            eg.
            la.image.filters(image, mode=la.image.FilterMode.EDGE_ENHANCE_MORE)

        la.image.FilterMode.EMBOSS:
            emboss blur.
            eg.
            la.image.filters(image, mode=la.image.FilterMode.EMBOSS)

        la.image.FilterMode.FIND_EDGES:
            Find the edge blur.
            eg.
            la.image.filters(image, mode=la.image.FilterMode.FIND_EDGES)

        la.image.FilterMode.SHARPEN:
            Sharpen blur.
            eg.
            la.image.filters(image, mode=la.image.FilterMode.SHARPEN)

        la.image.FilterMode.SMOOTH:
            Smooth blur.
            eg.
            la.image.filters(image, mode=la.image.FilterMode.SMOOTH)

        la.image.FilterMode.SMOOTH_MORE:
            Smooth threshold blur.
            eg.
            la.image.filters(image, mode=la.image.FilterMode.SMOOTH_MORE)

        Args:
            mode: la.image.FilterMode.
            p: probability that the image does this. Default value is 1.
        """
        if self._max_aug_nums>0:
            if self._nums>self._max_aug_nums:
                return self
            self._nums += 1
        if p is None:
            p = self._p
        self.image = filters(self.image, mode, p, **kwarg)
        return self



# import gc
import io
import wave

import av
import numpy as np

__all__ = ['read_audio', 'save_audio', 'AudioStream']


def save_audio(filename, audio_array, audio_fps=44100, audio_codec=None, audio_options=None):
    """Writes a 2d array in [C, L] format in a audio file.
    
    Args:
        filename: path where the video will be saved.
        audio_array: array[C, L] containing the audio, where C is the number of channels
            and L is the number of samples.
        audio_fps: audio sample rate, typically 44100 or 48000
        audio_codec: the name of the audio codec, i.e. "mp3", "aac", etc.
        audio_options: dictionary containing options to be passed into the PyAV audio stream
    """
    audio_format_dtypes = {
            "dbl": "<f8",
            "dblp": "<f8",
            "flt": "<f4",
            "fltp": "<f4",
            "s16": "<i2",
            "s16p": "<i2",
            "s32": "<i4",
            "s32p": "<i4",
            "u8": "u1",
            "u8p": "u1",
        }
    audio_codec_type = {'wav':'pcm_s16le'}
    if audio_codec is None:
        audio_codec = audio_codec_type.get(filename.split('.')[-1].lower(), filename.split('.')[-1].lower())
    if audio_codec in ['pcm_s16le']:
        container = wave.open(filename, "wb")
        container.setnchannels(len(audio_array.shape))
        container.setsampwidth(2)
        container.setframerate(int(audio_fps))
        container.setnframes(max(audio_array.shape))
        container.writeframes(audio_array.T.astype(np.int16).flat[:].tobytes())
    else:
        container = av.open(filename, mode="w")
        audio_layout = "stereo" if audio_array.shape[0] > 1 else "mono"
        stream = container.add_stream(audio_codec, rate=int(audio_fps))
        stream.options = audio_options or {}
        audio_sample_fmt = stream.codec_context.format.name
        dtype = np.dtype(audio_format_dtypes[audio_sample_fmt])
#         print(stream.codec_context.format.name, dtype)

#         frame = av.AudioFrame.from_ndarray(np.expand_dims(audio_array.T.flat[:].astype(dtype), 0),
#                                            format=audio_sample_fmt, layout=audio_layout)
        frame = av.AudioFrame.from_ndarray(audio_array.astype(dtype), 
                                           format=audio_sample_fmt, layout=audio_layout)
        frame.pts = None
        frame.sample_rate = int(audio_fps)

        for packet in stream.encode(frame):
            container.mux(packet)

        for packet in stream.encode():
            container.mux(packet)
    container.close()

# def save_audio(filename, audio, file_format=None):
#     """Saves an audio stored as a Numpy array to a path or file object.
        
#     Args
#         filename: Path or file object.
#         audio: a numpy array.
#         file_format: Optional file format override. If omitted, the
#             format to use is determined from the filename extension.
#     """
#     if file_format is None:
#         file_format = filename.split('.')[-1]
#     if file_format == "ogg":
#         file_format = 'libopus'#"libvorbis"
#     if file_format in ['WAV', 'wav']:
#         return save_wav(filename, audio.audio_data, params=audio.audio_params)
#     container = av.open(filename, 'w')
#     stream = container.add_stream(file_format)
#     frame = av.AudioFrame.from_ndarray(audio.audio_data, format='fltp')
#     frame.pts = None
#     frame.sample_rate = audio.audio_framerate
#     for packet in stream.encode(frame):
#         container.mux(packet)
#     container.close()

# def save_wav(filename, audio, params=None):
#     """Saves an audio stored as a Numpy array to a path or file object.
    
#     Now only support audio file format with '.wav'.
    
#     Args
#         filename: Path or file object.
#         audio: a numpy array.
#         params: {'audio_channel':None, 'audio_width': 2, 'audio_framerate': 44100, 'audio_frame':None}
#     """
#     param = {'audio_channel':None, 'audio_width': 2, 'audio_framerate': 44100, 'audio_frame':None}
#     if params is not None:
#         for i in params:
#             param[i] = params[i]
#     if param['audio_channel'] is None:
#         param['audio_channel'] = len(audio.shape)
#     if param['audio_frame'] is None:
#         param['audio_frame'] = max(audio.shape)
#     param['audio_frame'] = min(param['audio_frame'], max(audio.shape))
#     if filename[-4:] in ['.WAV', '.wav']:
#         f = wave.open(filename, "wb")
#         f.setnchannels(param['audio_channel'])
#         f.setsampwidth(param['audio_width'])
#         f.setframerate(param['audio_framerate'])
#         f.setnframes(param['audio_frame'])
#         f.writeframes(audio.T.astype(np.int16)[:param['audio_frame']].flat[:].tobytes())
#         f.close()
#     else:
#         raise ValueError(f'Not support audio file format with {filename}')
        

# class AudioWAV():
#     def __init__(self, filename):
#         """Reads the contents of file to a Audio instance.
        
#         Args:
#             filename: str, audio absolute path.
#         """
#         try:
#             self._file = wave.open(filename, 'rb')
#         except wave.Error:
#             raise ValueError('File is not a WAV file.')
            
#         self.audio_channel = self._file.getnchannels()
#         self.audio_width = self._file.getsampwidth()
#         self.audio_framerate = self._file.getframerate()
#         self.audio_frame = self._file.getnframes()
#         self.audio_duration = self.audio_frame / self.audio_framerate
#         self.audio_params = {'audio_channel':self.audio_channel, 'audio_width':self.audio_width,
#                              'audio_framerate':self.audio_framerate, 'audio_frame':self.audio_frame,
#                              'audio_duration':self.audio_duration,
#                              'audio_type':'wav', 'audio_name':filename[:-4]}
        
#         if self.audio_width not in (1, 2, 3, 4):
#             self.close()
#             raise ValueError('The file uses an unsupported bit width.')

#     def close(self):
#         """Close the underlying file."""
#         self._file.close()
        
#     @property
#     def audio_data(self):
#         """Reads and returns all frames of audio"""
#         data = self._get_data(self.audio_frame)
#         self._file.rewind()
#         return data
    
#     def stream(self, nframes=1024):
#         """Generates blocks of PCM data found in the file."""
#         while True:
#             data = self._get_data(nframes)
#             if not data:
#                 break
#             yield data
    
#     def _get_data(self, n):
#         str_data  = self._file.readframes(n)
#         return np.frombuffer(str_data, dtype=np.int16).reshape(-1,self.audio_channel).T

#     def __enter__(self):
#         return self

#     def __exit__(self, exc_type, exc_val, exc_tb):
#         self.close()
#         return False

#     def __iter__(self):
#         return self.stream()
    
    
# class Audio():
#     def __init__(self, filename='', data=None, audio_params=None):
#         """Reads the contents of file to a Vedio instance.
        
#         Args:
#             filename: str, vedio absolute path.
#         """
#         self._data = data
#         self.audio_params = audio_params
#         if filename:
#             self._file = av.open(filename, 'r')
#             audio = self._file.streams.audio[0]

#             self.audio_size      = self._file.size
#             self.audio_channel   = audio.codec_context.channels
#     #         self.audio_width = self._file.getsampwidth()
#             self.audio_framerate = audio.codec_context.sample_rate
#             self.audio_frame     = audio.duration
#             self.audio_duration  = self._file.duration/1000000
# #             self.close()
#         self.audio_params = {'audio_channel':self.audio_channel, 
#                             'audio_duration':self.audio_duration,
#                             'audio_framerate':self.audio_framerate, 
#                             'audio_frame':self.audio_frame,
#                             'audio_size':self.audio_size, 'audio_path':filename}
        
#     def close(self):
#         """Close the underlying file."""
#         self._file.close()

#     @property
#     def audio_data(self):
#         """Reads and returns all frames of audio"""
#         if self._data is None:
#             with av.open(self.audio_params['audio_path'], 'r') as container:
#                 self._data = np.concatenate([frame.to_ndarray() for frame in container.decode(audio=0)], axis=1)
#         return self._data.reshape(-1,self.audio_channel).T
    
#     def stream(self):
#         """Generates blocks of PCM data found in the file."""
#         if self._data is None:
#             with av.open(self.audio_params['audio_path'], 'r') as container:
#                 self._data = [frame.to_ndarray() for frame in container.decode(audio=0)]
#         for frame in self._data:
#             yield frame.reshape(-1,self.audio_channel).T
    
#     def __enter__(self):
#         return self

#     def __exit__(self, exc_type, exc_val, exc_tb):
#         self.close()
#         return False

#     def __iter__(self):
#         return self.stream()


    
def read_audio(src, start_sec=0, end_sec=np.inf, audio_format="CL", **kwargs):
    """Reads a audio from a file, returning both the audio frames and audio metadata.
    
    Args:
        src: path to the audio file or bytes.
        start_sec: float, The start presentation time of the audio.
        end_sec: float, The end presentation time of the audio.
        audio_format: The format of the output audio shape. 'CL' or 'LC' e.g. 
    Returns:
        audio: array[C, L]): the audio frames, where `C` is the number of channels and `L` is the number of points.
        metadata: metadata for the audio.
    """
    audio_format = audio_format.upper()
    if len([i for i in set(audio_format) if i in 'CL'])!=len(audio_format):
        raise ValueError(f"`data_format` should be 'CL' or 'LC' e.g., got {audio_format}.")

    if end_sec < start_sec:
        raise ValueError("end_sec should be larger than start_sec")

    if isinstance(src, bytes):
        src = io.BytesIO(src)
        info = {'filename':''}
    else:
        info = {'filename':src}

    if info['filename'].split('.')[-1].lower()=='wav':
        container = wave.open(info['filename'], 'rb')
        info["audio_fps"]       = container.getframerate()
        info["audio_channel"]   = container.getnchannels()
        info["audio_frames"]    = container.getnframes()
        info["audio_duration"]  = info["audio_frames"]/info["audio_fps"]
        str_data = container.readframes(container.getnframes())
        aframes = np.frombuffer(str_data, dtype=np.float32).reshape(-1, container.getnchannels()).T
        container.rewind()
        start_time = 0
    else:
        container = av.open(src, metadata_errors="ignore")
        info["audio_fps"]       = container.streams.audio[0].rate
        info["audio_channel"]   = container.streams.audio[0].codec_context.channels
        aframes = [frame.to_ndarray() for frame in container.decode(audio=0)]
        if aframes:
            aframes = np.concatenate(aframes, axis=1, dtype=np.float32)
            if aframes.shape[0]!=info["audio_channel"]:
                aframes = aframes.reshape(-1, info["audio_channel"]).T
        else:
            aframes = np.empty((info["audio_channel"], 0), dtype=np.float32)
        info["audio_frames"]    = container.streams.audio[0].duration
        if info["audio_frames"] is None:
            info["audio_frames"] = aframes.shape[1]
        info["audio_duration"]  = info["audio_frames"]*float(container.streams.audio[0].time_base)
        start_time = container.streams.audio[0].start_time/1000000
    
    container.close()
#     gc.collect()

    interval = float(info["audio_duration"])-start_time
    batch_start_time = max(start_sec, start_time)
    batch_end_time = min(end_sec, float(info["audio_duration"]))
    start = (batch_start_time-start_time)/interval
    end = (batch_end_time-start_time)/interval
    aframes = aframes[:, int(aframes.shape[1]*start):int(aframes.shape[1]*end)]

    if 'batch' in kwargs:
        batch_time = kwargs['batch']/aframes.shape[1]*interval
        return {'audio':aframes, 'metadata':info, 'batch_time':batch_time,
                'batch_start_time':batch_start_time, 'batch_end_time':batch_end_time}
    if audio_format!='CL':
        transpose = {'C':0, 'L':1}
        aframes = aframes.transpose(tuple(transpose[i] for i in audio_format))
    return {'audio':aframes, 'metadata':info}


class AudioStream():
    def __init__(self, src, batch=4096, data_format="CL", start_sec=0, end_sec=np.inf):
        """Converts a Audio instance to a Numpy array.
        
        Args:
            src: input audio path or bytes.
            batch: each iter batch number.
            data_format: array data format, 'CL' or 'LC' e.g. 
            start_sec: float, The start presentation time of the audio.
            end_sec: float, The end presentation time of the audio.
        Returns:
            A Numpy array iterator.
        """
        data = read_audio(src, start_sec=start_sec, end_sec=end_sec, audio_format=data_format, batch=batch)
        self._batch            = int(batch)
        self._data_format      = data_format.upper()
        self._aframes          = data['audio']
        self.metadata          = data['metadata']
        self._batch_time       = data['batch_time']
        self._batch_start_time = data['batch_start_time']
        self._batch_end_time   = data['batch_end_time']
        
    def __next__(self):
        try:
            batch_pts = []
            count = 0
            if not self._aframes.shape[1]:
                raise StopIteration
            batch_array = self._aframes[:,:self._batch]
            self._aframes = self._aframes[:,self._batch:]
            batch_pts.append(self._batch_start_time)
            if batch_array.shape[1]<self._batch:
                batch_pts.append(self._batch_end_time)
            else:
                self._batch_start_time += self._batch_time
                batch_pts.append(self._batch_start_time)
            if self._data_format!='CL':
                transpose = {'C':0, 'L':1}
                batch_array = batch_array.transpose(tuple(transpose[i] for i in self._data_format))
        except av.error.EOFError:
            raise StopIteration
        return {"data": batch_array, "pts": batch_pts}

    def __iter__(self):
        return self




import numpy as np
import matplotlib.pyplot as plt

from linora.chart._config import Options
from linora.chart._base import Coordinate
from linora.chart._arrow import Arrow
from linora.chart._bar import Bar
from linora.chart._boxplot import Boxplot
from linora.chart._circle import Circle
from linora.chart._coherence import Coherence
from linora.chart._csd import Csd
from linora.chart._ellipse import Ellipse
from linora.chart._errorbar import Errorbar
from linora.chart._fillline import Fillline
from linora.chart._hist import Hist
from linora.chart._hist2d import Hist2d
from linora.chart._hlines import Hlines
from linora.chart._line import Line
from linora.chart._line3D import Line3D
from linora.chart._pie import Pie
from linora.chart._polygon import Polygon
from linora.chart._radar import Radar
from linora.chart._rectangle import Rectangle
from linora.chart._regularpolygon import RegularPolygon
from linora.chart._scatter import Scatter
from linora.chart._scatter3D import Scatter3D
from linora.chart._vlines import Vlines
from linora.chart._wedge import Wedge


__all__ = ['Plot']


classlist = [
    Coordinate, Arrow, Bar, Boxplot, Circle, Coherence, Csd, Ellipse, Errorbar, Fillline, 
    Hist, Hist2d, Hlines, Line, Line3D,
    Pie, Polygon, Radar, Rectangle, RegularPolygon, Scatter, Scatter3D, Vlines, Wedge
]

class Plot(*classlist):
    def __init__(self, config=None):
        super(Plot, self).__init__()
        if config is not None:
            self.set_config(config)
        key = np.random.choice(list(Options.color), size=len(Options.color), replace=False)
        t = {i:np.random.choice(Options.color[i], size=len(Options.color[i]), replace=False) for i in key}
        self._params.color = sorted([[r+k*7, j, i] for r, i in enumerate(key) for k, j in enumerate(t[i])])
#         if len(args)!=0:
#             if isinstance(args[0], dict):
#                 for i,j in args[0].items():
#                     setattr(self._params, i, j)
#         if kwargs:
#             for i,j in kwargs.items():
#                 setattr(self._params, i, j)
                
    def _execute(self):
        fig = plt.figure(**self._params.figure)
        with plt.style.context(self._params.theme):
            mode = set([j['plotmode'].split('_')[-1] if '_' in j['plotmode'] else 'rectilinear' for i,j in self._params.ydata.items()])
            if len(mode)==1:
                projection = list(mode)[0]
            elif len(mode)==2 and 'rectilinear' in mode and '3d' in mode:
                projection = '3d'
            else:
                raise ValueError('There are two different coordinate systems.')
            ax = fig.add_subplot(projection=projection)
            
            for i,j in self._params.ydata.items():
                if 'transform' in j['kwargs']:
                    if j['kwargs']['transform'] in ['transData', 'Data']:
                        j['kwargs']['transform'] = ax.transData
                    elif j['kwargs']['transform'] in ['transAxes', 'Axes']:
                        j['kwargs']['transform'] = ax.transAxes
                    elif j['kwargs']['transform'] in ['transFigure', 'Figure']:
                        j['kwargs']['transform'] = fig.transFigure
                        j['transform'] = 'fig'
            
        ax = self._execute_ax(fig, ax)
        return fig
            
    def _execute_ax(self, fig, ax):
        for i,j in self._params.ydata.items():
            if i in self._params.twin:
                if self._params.twin[i]=='x':
                    ax_new = ax.twinx()
                    j['plotfunc'](fig, ax_new, i, j)
                    if 'x' in self._params.label:
                        self._execute_label(ax_new, 'x')
                    if 'x' in self._params.axis:
                        self._execute_axis(ax_new, 'x')
                else:
                    ax_new = ax.twiny()
                    j['plotfunc'](fig, ax_new, i, j)
                    if 'y' in self._params.label:
                        self._execute_label(ax_new, 'y')
                    if 'y' in self._params.axis:
                        self._execute_axis(ax_new, 'y')
            else:
                j['plotfunc'](fig, ax, i, j)
                self._execute_label(ax, 'normal')
                self._execute_axis(ax, 'normal')
            
        #legend
        if len(self._params.legend)>0:
            if self._params.legend['loc'] not in [None, 'None', 'none']:
                ax.legend(**self._params.legend)
        else:
            t = ['ellipse', 'regularpolygon', 'rectangle', 'circle', 'polygon', 'boxplot', 'wedge', 'arrow']
            if len([1 for i,j in self._params.ydata.items() if j['plotmode'] not in t])>1:
                ax.legend(loc='best')
        #spines
        if len(self._params.spine['alpha'])>0:
            for i,j in self._params.spine['alpha'].items():
                ax.spines[i].set_alpha(j)
        if len(self._params.spine['color'])>0:
            for i,j in self._params.spine['color'].items():
                ax.spines[i].set_color(j)
        if len(self._params.spine['width'])>0:
            for i,j in self._params.spine['width'].items():
                ax.spines[i].set_linewidth(j)
        if len(self._params.spine['style'])>0:
            for i,j in self._params.spine['style'].items():
                ax.spines[i].set_linestyle(j)
        if len(self._params.spine['position'])>0:
            for i,j in self._params.spine['position'].items():
                ax.spines[i].set_position(j)
        if len(self._params.spine['show'])>0:
            for i,j in self._params.spine['show'].items():
                ax.spines[i].set_visible(j)
        #title
        if self._params.title['label'] is not None:
            ax.set_title(**self._params.title)
        #text
        if len(self._params.text)>0:
            for i in self._params.text:
                ax.text(**self._params.text[i])
        #annotate
        if len(self._params.annotate)>0:
            for i in self._params.annotate:
                ax.annotate(**self._params.annotate[i])
        return ax
    
    def _execute_label(self, ax, mode):
        if self._params.label[mode]['xlabel']['xlabel'] is not None:
            ax.set_xlabel(**self._params.label[mode]['xlabel'])
        if self._params.label[mode]['ylabel']['ylabel'] is not None:
            ax.set_ylabel(**self._params.label[mode]['ylabel'])
            
    def _execute_axis(self, ax, mode):
        if self._params.axis[mode]['axis'] is not None:
            ax.axis(self._params.axis[mode]['axis'])
        if self._params.axis[mode]['xlabel'] is not None:
            if len(self._params.axis[mode]['xlabel'])==0:
                ax.set_xticks(self._params.axis[mode]['xlabel'])
            elif isinstance(self._params.axis[mode]['xlabel'][0], (list, tuple, np.ndarray)):
                ax.set_xticks(self._params.axis[mode]['xlabel'][0])
                ax.set_xticklabels(self._params.axis[mode]['xlabel'][1])
            else:
                ax.set_xticks(self._params.axis[mode]['xlabel'])
        if self._params.axis[mode]['ylabel'] is not None:
            if len(self._params.axis[mode]['ylabel'])==0:
                ax.set_yticks(self._params.axis[mode]['ylabel'])
            elif isinstance(self._params.axis[mode]['ylabel'][0], (list, tuple, np.ndarray)):
                ax.set_yticks(self._params.axis[mode]['ylabel'][0])
                ax.set_yticklabels(self._params.axis[mode]['ylabel'][1])
            else:
                ax.set_yticks(self._params.axis[mode]['ylabel'])
        ax.tick_params(axis='x', **self._params.axis[mode]['xtick'])
        ax.tick_params(axis='y', **self._params.axis[mode]['ytick'])
        if self._params.axis[mode]['xinvert']:
            ax.invert_xaxis()
        if self._params.axis[mode]['yinvert']:
            ax.invert_yaxis()
        if self._params.axis[mode]['xtickposition'] is not None:
            ax.xaxis.set_ticks_position(self._params.axis[mode]['xtickposition'])
        if self._params.axis[mode]['ytickposition'] is not None:
            ax.yaxis.set_ticks_position(self._params.axis[mode]['ytickposition'])


import sys
import time
from collections import Counter

import numpy as np
import pandas as pd
from linora.sample_splits import kfold, train_test_split


def RandomSearch(feature, label, estimator, params, metrics, iter_num=1000, scoring=0.5, cv=5, cv_num=3,
                 metrics_min=True, speedy=True, speedy_param=(20000, 0.3), stratify=None):
    """General model params search use RandomSearch method.
    
    Args:
        feature: pandas dataframe, model's feature.
        label: pandas series, model's label.
        estimator: general model.
        params: a general model params dict.
        metrics: model metrics function.
        scoring: metrics error opt base line value.
        cv: cross validation fold.
        cv_num: minimum cross validation fold.
        metrics_min: metrics value whether the smaller the better.
        speedy: whether use speedy method.
        speedy_param: if use speedy method, test_size will be set, 
                      test_size = 1-round(min(speedy_param[0], feature.shape[0]*speedy_param[1])/feature.shape[0], 2).
        gpu: whether use gpu.
    Returns:
        a best General model params dict.
    Raises:
        params error.
    """
    start = time.time()
    best_params={}
    if speedy:
        test_size = round(min(speedy_param[0], feature.shape[0]*speedy_param[1])/feature.shape[0], 2)
    for i in range(1, iter_num+1):
        param = {i:np.random.choice(j) for i,j in params.items()}
        model = estimator(**param)
        score = []
        if speedy:
            for _ in range(cv_num):
                index_list = train_test_split(feature, stratify=stratify, test_size=test_size,
                                              shuffle=True, random_state=np.random.choice(range(100), 1)[0])
                model.fit(feature.loc[index_list[0]], label[index_list[0]])
                cv_pred = pd.Series(model.predict(feature.loc[index_list[1]]), index=label[index_list[1]].index)
                score.append(metrics(label[index_list[1]], cv_pred))
        else:
            index_list = kfold(feature, stratify=stratify, n_splits=cv, shuffle=True, random_state=np.random.choice(range(100), 1)[0])
            for n, index in enumerate(index_list):
                if n == cv_num:
                    break
                model.fit(feature.loc[index[0]], label[index[0]])
                cv_pred = pd.Series(model.predict(feature.loc[index[1]]), index=label[index[1]].index)
                score.append(metrics(label[index[1]], cv_pred))
        cv_score = round(np.mean(score), 4)
        if metrics_min:
            if cv_score<scoring:
                scoring = cv_score
                best_params = param.copy()
        else:
            if cv_score>scoring:
                scoring = cv_score
                best_params = param.copy()
        sys.stdout.write("GERegressor random search percent: {}%, run time {} min, best score: {}, best param：{}\r".format(
            round(i/iter_num*100,2), divmod((time.time()-start),60)[0], scoring, best_params))
        sys.stdout.flush()
    print("GERegressor param finetuning with random search run time: %d min %.2f s" % divmod((time.time() - start), 60))
    return best_params



import requests

__all__ = ['BotFeiShu']


class BotFeiShu():
    """send message by Feishu
    
    Args:
        webhook: Feishu robot url
    """
    def __init__(self, webhook):
        self.webhook = webhook
        
    def send_text(self, msg):
        """send message
        
        Args:
            msg: str, message
        Return:
            request post information.
        """
        t = requests.post(self.webhook, json={"msg_type": "text", "content":{"text": msg}})
        return t.content.decode()


import numpy as np

__all__ = ['normalize_max', 'normalize_maxabs', 'normalize_l1', 'normalize_l2', 
           'normalize_meanminmax', 'normalize_minmax', 'normalize_norm', 'normalize_robust']


def normalize_max(feature, mode=0, name=None, config=None):
    """normalize feature with max method.
    
    Args:
        feature: pd.Series, sample feature value.
        mode: if 0, output (transform feature, config); if 1, output transform feature; if 2, output config.
        name: str, output feature name, if None, name is feature.name .
        config: dict, label parameters dict for this estimator. 
            if config is not None, only parameter `feature` and `mode` is invalid.
    Returns:
        Refer to params `mode` explanation.
    """
    if config is None:
        config = {'param':{'feature_scale':feature.max(), 'name':feature.name if name is None else name},
                  'type':'normalize_max', 'variable':feature.name}
    if mode==2:
        return config
    else:
        scale = config['param']['feature_scale']
        t = (feature/scale).rename(config['param']['name'])
        return t if mode else (t, config)
    

def normalize_maxabs(feature, mode=0, name=None, config=None):
    """normalize feature with maxabs method.
    
    Args:
        feature: pd.Series, sample feature value.
        mode: if 0, output (transform feature, config); if 1, output transform feature; if 2, output config.
        name: str, output feature name, if None, name is feature.name .
        config: dict, label parameters dict for this estimator. 
            if config is not None, only parameter `feature` and `mode` is invalid.
    Returns:
        Refer to params `mode` explanation.
    """
    if config is None:
        config = {'param':{'feature_scale':feature.abs().max(), 'name':feature.name if name is None else name},
                  'type':'normalize_maxabs', 'variable':feature.name}
    if mode==2:
        return config
    else:
        scale = config['param']['feature_scale']
        t = (feature/scale).rename(config['param']['name'])
        return t if mode else (t, config)


def normalize_l1(feature, mode=0, name=None, config=None):
    """normalize feature with l1 method.
    
    Args:
        feature: pd.Series, sample feature value.
        mode: if 0, output (transform feature, config); if 1, output transform feature; if 2, output config.
        name: str, output feature name, if None, name is feature.name .
        config: dict, label parameters dict for this estimator. 
            if config is not None, only parameter `feature` and `mode` is invalid.
    Returns:
        Refer to params `mode` explanation.
    """
    if config is None:
        config = {'param':{'feature_scale':feature.abs().sum(), 'name':feature.name if name is None else name},
                  'type':'normalize_l1', 'variable':feature.name}
    if mode==2:
        return config
    else:
        scale = config['param']['feature_scale']
        t = (feature/scale).rename(config['param']['name'])
        return t if mode else (t, config)


def normalize_l2(feature, mode=0, name=None, config=None):
    """normalize feature with l2 method.
    
    Args:
        feature: pd.Series, sample feature value.
        mode: if 0, output (transform feature, config); if 1, output transform feature; if 2, output config.
        name: str, output feature name, if None, name is feature.name .
        config: dict, label parameters dict for this estimator. 
            if config is not None, only parameter `feature` and `mode` is invalid.
    Returns:
        Refer to params `mode` explanation.
    """
    if config is None:
        config = {'param':{'feature_scale':np.sqrt(np.sum(np.square(feature))), 
                           'name':feature.name if name is None else name},
                  'type':'normalize_l2', 'variable':feature.name}
    if mode==2:
        return config
    else:
        scale = config['param']['feature_scale']
        t = (feature/scale).rename(config['param']['name_output'])
        return t if mode else (t, config)


def normalize_meanminmax(feature, mode=0, name=None, config=None):
    """normalize feature with meanminmax method.
    
    Args:
        feature: pd.Series, sample feature value.
        mode: if 0, output (transform feature, config); if 1, output transform feature; if 2, output config.
        name: str, output feature name, if None, name is feature.name .
        config: dict, label parameters dict for this estimator. 
            if config is not None, only parameter `feature` and `mode` is invalid.
    Returns:
        Refer to params `mode` explanation.
    """
    if config is None:
        config = {'param':{'feature_scale':[feature.mean(), feature.min(), feature.max()],
                           'name':feature.name if name is None else name},
                  'type':'normalize_meanminmax', 'variable':feature.name}
    if mode==2:
        return config
    else:
        scale = config['param']['feature_scale']
        t = ((feature-scale[0])/(scale[2]-scale[1])).rename(config['param']['name'])
        return t if mode else (t, config)


def normalize_minmax(feature, mode=0, feature_range=(0, 1), name=None, config=None):
    """normalize feature with minmax method.
    
    Args:
        feature: pd.Series, sample feature value.
        mode: if 0, output (transform feature, config); if 1, output transform feature; if 2, output config.
        feature_range: list or tuple, range of values after feature transformation.
        name: str, output feature name, if None, name is feature.name .
        config: dict, label parameters dict for this estimator. 
            if config is not None, only parameter `feature` and `mode` is invalid.
    Returns:
        Refer to params `mode` explanation.
    """
    if config is None:
        config = {'param':{'feature_scale':[feature.min(), feature.max()], 
                           'feature_range':feature_range, 'name':feature.name if name is None else name},
                  'type':'normalize_minmax', 'variable':feature.name}
    if mode==2:
        return config
    else:
        scale = config['param']['feature_scale']
        ran = config['param']['feature_range']
        t = ((feature-scale[0])/(scale[1]-scale[0])*(ran[1]-ran[0])+ran[0]).rename(config['param']['name'])
        return t if mode else (t, config)


def normalize_norm(feature, mode=0, name=None, config=None):
    """normalize feature with norm method.
    
    Args:
        feature: pd.Series, sample feature value.
        mode: if 0, output (transform feature, config); if 1, output transform feature; if 2, output config.
        name: str, output feature name, if None, name is feature.name .
        config: dict, label parameters dict for this estimator. 
            if config is not None, only parameter `feature` and `mode` is invalid.
    Returns:
        Refer to params `mode` explanation.
    """
    if config is None:
        config = {'param':{'feature_scale':[feature.mean(), feature.std()],
                           'name':feature.name if name is None else name},
                  'type':'normalize_norm', 'variable':feature.name}
    if mode==2:
        return config
    else:
        scale = config['param']['feature_scale']
        t = ((feature-scale[0])/scale[1]).rename(config['param']['name'])
        return t if mode else (t, config)


def normalize_robust(feature, mode=0, feature_scale=(0.5, 0.5), name=None, config=None):
    """normalize feature with robust method.
    
    Args:
        feature: pd.Series, sample feature value.
        mode: if 0, output (transform feature, config); if 1, output transform feature; if 2, output config.
        feature_scale: list or tuple, each element is in the [0,1] interval.
                       (feature_scale[0], feature.quantile(0.5+feature_scale[1]/2)-feature.quantile(0.5-feature_scale[1]/2)).
        name: str, output feature name, if None, name is feature.name .
        config: dict, label parameters dict for this estimator. 
            if config is not None, only parameter `feature` and `mode` is invalid.
    Returns:
        Refer to params `mode` explanation.
    """
    if config is None:
        config = {'param':{'feature_scale':[feature.quantile(feature_scale[0]), feature.quantile(0.5+feature_scale[1]/2)-feature.quantile(0.5-feature_scale[1]/2)], 
                           'name':feature.name if name is None else name},
                  'type':'normalize_robust', 'variable':feature.name}
    if mode==2:
        return config
    else:
        scale = config['param']['feature_scale']
        t = ((feature-scale[0])/scale[1]).rename(config['param']['name'])
        return t if mode else (t, config)



from PIL import ImageChops

__all__ = ['pairs', 'PairsMode']


class pairs_mode:
    """pairs image computing method type."""
    Add               = 'add'
    Add_modulo        = 'add_modulo'
    Blend             = 'blend'
    Composite         = 'composite'
    Darker            = 'darker'
    Difference        = 'difference'
    Lighter           = 'lighter'
    Logical_and       = 'logical_and'
    Logical_or        = 'logical_or'
    Logical_xor       = 'logical_xor'
    Multiply          = 'multiply'
    SoftLight         = 'soft_light'
    HardLight         = 'hard_light'
    Overlay           = 'overlay'
    Screen            = 'screen'
    Subtract          = 'subtract'
    Subtract_modulo   = 'subtract_modulo'

    
PairsMode = pairs_mode()


def pairs(image1, image2, mode, **kwargs):
    """
    la.image.PairsMode.Add
        Adds two images, dividing the result by scale and adding the offset. If omitted, scale defaults to 1.0, and offset to 0.0.
        out = ((image1 + image2) / scale + offset)
        
        you should append param `scale` and `offset`
        scale: int or float, defaults to 1.0
        offset: int or float, defaults to 0
        eg.
        la.image.pairs(image1, image2, mode=la.image.PairsMode.Add, scale=1.0, offset=0)
    
    la.image.PairsMode.Add_modulo
        Add two images, without clipping the result.
        out = ((image1 + image2) % 255)
        eg.
        la.image.pairs(image1, image2, mode=la.image.PairsMode.Add_modulo)
        
    la.image.PairsMode.Blend
        Creates a new image by interpolating between two input images.

        using a constant alpha.
        out = image1 * (1.0 - alpha) + image2 * alpha

        If alpha is 0.0, a copy of the first image is returned. 
        If alpha is 1.0, a copy of the second image is returned. 
        There are no restrictions on the alpha value. 
        If necessary, the result is clipped to fit into the allowed output range.
        
        you should append param `alpha`
        alpha: The interpolation alpha factor.
        eg.
        la.image.pairs(image1, image2, mode=la.image.PairsMode.Blend, alpha=0.2)
        
    la.image.PairsMode.Composite
        Create composite image by blending images using a transparency mask.
        
        you should append param `mask`
        mask: A mask image. This image can have mode “1”, “L”, or “RGBA”, and must have the same size as the other two images.
        eg.
        la.image.pairs(image1, image2, la.image.PairsMode.Composite, mask)
        
    la.image.PairsMode.Darker
        Compares the two images, pixel by pixel, and returns a new image containing the darker values.
        out = min(image1, image2)
        
        eg.
        la.image.pairs(image1, image2, mode=la.image.PairsMode.Darker)
        
    la.image.PairsMode.Difference
        Returns the absolute value of the pixel-by-pixel difference between the two images.
        out = abs(image1 - image2)
        
        eg.
        la.image.pairs(image1, image2, mode=la.image.PairsMode.Difference)
    
    la.image.PairsMode.Lighter
        Compares the two images, pixel by pixel, and returns a new image containing the lighter values.
        out = max(image1, image2)
        
        eg.
        la.image.pairs(image1, image2, mode=la.image.PairsMode.Lighter)
        
    la.image.PairsMode.Logical_and
        Logical AND between two images.
        Both of the images must have mode “1”. 
        If you would like to perform a logical AND on an image with a mode other than “1”, 
        try multiply() instead, using a black-and-white mask as the second image.
        out = ((image1 and image2) % 255)
        
        eg.
        la.image.pairs(image1, image2, mode=la.image.PairsMode.Logical_and)
        
    la.image.PairsMode.Logical_or
        Logical OR between two images.
        Both of the images must have mode “1”.
        out = ((image1 or image2) % 255)
        
        eg.
        la.image.pairs(image1, image2, mode=la.image.PairsMode.Logical_or)
        
    la.image.PairsMode.Logical_xor
        Logical XOR between two images.
        Both of the images must have mode “1”.
        out = ((bool(image1) != bool(image2)) % 255)
        
        eg.
        la.image.pairs(image1, image2, mode=la.image.PairsMode.Logical_xor)
        
    la.image.PairsMode.multiply
        Superimposes two images on top of each other.
        If you multiply an image with a solid black image, the result is black. 
        If you multiply with a solid white image, the image is unaffected.
        out = image1 * image2 / 255
        
        eg.
        la.image.pairs(image1, image2, mode=la.image.PairsMode.multiply)
        
    la.image.PairsMode.SoftLight
        Superimposes two images on top of each other using the Soft Light algorithm
        eg.
        la.image.pairs(image1, image2, mode=la.image.PairsMode.SoftLight)
        
    la.image.PairsMode.HardLight
        Superimposes two images on top of each other using the Hard Light algorithm
        eg.
        la.image.pairs(image1, image2, mode=la.image.PairsMode.HardLight)
        
    la.image.PairsMode.Overlay
        Superimposes two images on top of each other using the Overlay algorithm
        eg.
        la.image.pairs(image1, image2, mode=la.image.PairsMode.Overlay)
        
    la.image.PairsMode.Screen
        Superimposes two inverted images on top of each other.
        out = 255 - ((255 - image1) * (255 - image2) / 255)
        eg.
        la.image.pairs(image1, image2, mode=la.image.PairsMode.Screen)
        
    la.image.PairsMode.Subtract
        Subtracts two images, dividing the result by scale and adding the offset. If omitted, scale defaults to 1.0, and offset to 0.0.
        out = ((image1 - image2) / scale + offset)
        
        you should append param `scale` and `offset`
        scale: int or float, defaults to 1.0
        offset: int or float, defaults to 0
        eg.
        la.image.pairs(image1, image2, mode=la.image.PairsMode.Subtract, scale=1.0, offset=0)
        
    la.image.PairsMode.Subtract_modulo
        Subtract two images, without clipping the result.
        out = ((image1 - image2) % 255)
        eg.
        la.image.pairs(image1, image2, mode=la.image.PairsMode.Subtract_modulo)


    Args:
        image1: a PIL instance. The first image.
        image2: a PIL instance. The second image.  Must have the same mode and size as the first image.
        mode: la.image.PairsMode
    Return:
        a PIL instance.
    """
    if 'scale' not in kwargs:
        kwargs['scale'] = 1.
    if 'offset' not in kwargs:
        kwargs['offset'] = 0
    if mode=='add':
        return ImageChops.add(image1, image2, scale=kwargs['scale'], offset=kwargs['offset'])
    elif mode=='add_modulo':
        return ImageChops.add_modulo(image1, image2)
    elif mode=='blend':
        if 'alpha' not in kwargs:
            raise ValueError("Missing parameter `alpha`")
        return ImageChops.blend(image1, image2, alpha=kwargs['alpha'])
    elif mode=='composite':
        if 'mask' not in kwargs:
            raise ValueError("Missing parameter `mask`")
        return ImageChops.composite(image1, image2, mask=kwargs['mask'])
    elif mode=='darker':
        return ImageChops.darker(image1, image2)
    elif mode=='difference':
        return ImageChops.difference(image1, image2)
    elif mode=='lighter':
        return ImageChops.lighter(image1, image2)
    elif mode=='logical_and':
        return ImageChops.logical_and(image1, image2)
    elif mode=='logical_or':
        return ImageChops.logical_or(image1, image2)
    elif mode=='logical_xor':
        return ImageChops.logical_xor(image1, image2)
    elif mode=='multiply':
        return ImageChops.multiply(image1, image2)
    elif mode=='soft_light':
        return ImageChops.soft_light(image1, image2)
    elif mode=='hard_light':
        return ImageChops.hard_light(image1, image2)
    elif mode=='overlay':
        return ImageChops.overlay(image1, image2)
    elif mode=='screen':
        return ImageChops.screen(image1, image2)
    elif mode=='subtract':
        return ImageChops.subtract(image1, image2, scale=kwargs['scale'], offset=kwargs['offset'])
    elif mode=='subtract_modulo':
        return ImageChops.subtract_modulo(image1, image2)
    else:
        raise ValueError("mode must be la.image.PairsMode param")


import itertools

import numpy as np
from PIL import Image

from linora.image._image_draw import draw_box
from linora.image._image_feature import histogram

__all__ = ['box_area', 'box_convert', 'mask_to_box']


def box_area(box):
    """Computes the area of a set of bounding boxes.
    
    Args:
        box: boxes for which the area will be computed. eg.(x1, y1, x2, y2,...) or [[x1, y1], [x2, y2], ...]
    Returns:
        the area for box.
    """
    if isinstance(box[0], (list, tuple)):
        box = list(itertools.chain.from_iterable(box))
    image = Image.new('L', [max([box[i*2] for i in range(int(len(box)/2))]), max([box[i*2+1] for i in range(int(len(box)/2))])])
    image = draw_box(image, box, fill_color=255, line_color=255)
    return histogram(image, if_global=True)[-1]


def mask_to_box(mask):
    """Compute the bounding boxes around the provided masks.
    
    Args:
        mask: a numpy array, shape is (H, W).
    returns: 
        a list of bounding boxes, eg [[x1,y1,x2,y2],...].
    """
    axis = []
    for i in np.unique(mask):
        if i!=0:
            y, x = np.where(mask==i)
            axis.append([x.min(), y.min(), x.max(), y.max()])
    return axis


def box_convert(box, in_fmt, out_fmt):
    """Converts boxes from given in_fmt to out_fmt. 
    
    Supported in_fmt and out_fmt are:
    'xyxy': boxes are represented via corners, x1, y1 being top left and x2, y2 being bottom right. 
    'xywh': boxes are represented via corner, width and height, x1, y1 being top left, w, h being width and height.
    'cxcywh': boxes are represented via centre, width and height, cx, cy being center of box, w, h being width and height.
    'axis': all coordinates of the 4 points in the boxes.
    Args:
        box: 4 tuple or list, boxes which will be converted.
        in_fmt: str, input format of given boxes. Supported formats are ['xyxy', 'xywh', 'cxcywh', 'axis'].
        out_fmt: str, output format of given boxes. Supported formats are ['xyxy', 'xywh', 'cxcywh', 'axis'].
    Returns:
        a box, list, boxes which will be converted.
    """
    if in_fmt=='xyxy' and out_fmt=='xywh':
        return [box[0], box[1], box[2]-box[0], box[3]-box[1]]
    if in_fmt=='xyxy' and out_fmt=='cxcywh':
        return [(box[2]+box[0])/2, (box[3]+box[1])/2, box[2]-box[0], box[3]-box[1]]
    if in_fmt=='xyxy' and out_fmt=='axis':
        return [box[0], box[1], box[2], box[1], box[2], box[3], box[0], box[3]]
    if in_fmt=='xywh' and out_fmt=='xyxy':
        return [box[0], box[1], box[2]+box[0], box[3]+box[1]]
    if in_fmt=='xywh' and out_fmt=='cxcywh':
        return [box[0]+box[2]/2, box[1]+box[3]/2, box[2], box[3]]
    if in_fmt=='xywh' and out_fmt=='axis':
        return [box[0], box[1], box[0]+box[2], box[1], box[0]+box[2], box[1]+box[3], box[0], box[1]+box[3]]
    if in_fmt=='cxcywh' and out_fmt=='xywh':
        return [box[0]-box[2]/2, box[1]-box[3]/2, box[2], box[3]]
    if in_fmt=='cxcywh' and out_fmt=='xyxy':
        return [box[0]-box[2]/2, box[1]-box[3]/2, box[0]+box[2]/2, box[1]+box[3]/2]
    if in_fmt=='cxcywh' and out_fmt=='axis':
        return [box[0]-box[2]/2, box[1]-box[3]/2, box[0]+box[2]/2, box[1]-box[3]/2, 
                box[0]+box[2]/2, box[1]+box[3]/2, box[0]-box[2]/2, box[1]+box[3]/2]
    if in_fmt=='axis' and out_fmt=='xyxy':
        return [box[0], box[1], box[4], box[5]]
    if in_fmt=='axis' and out_fmt=='cxcywh':
        return [(box[0]+box[4])/2, (box[1]+box[5])/2, box[4]-box[0], box[5]-box[1]]
    if in_fmt=='axis' and out_fmt=='xywh':
        return [box[0], box[1], box[4]-box[0], box[5]-box[1]]


import time
import datetime
import threading
from collections import defaultdict

__all__ = ['ThreadLoom']


class Params:
    pass
    
    
class ThreadLoom():
    """ThreadLoom class: executes runners using threading."""
    def __init__(self, max_runner, mode=1):
        """
        max_runner: int, The total number of runners that are allowed to be running at any given time.
        mode: 1 is setDaemon(True) and 0 is join().
        """
        self.params = Params()
        self.params.max_runner = max_runner
        self.params.mode = mode
        self.params.runners = list()
        self.params.started = list()
        self.params.time_pause = 0.1
        self.params.tracker_dict = defaultdict()
        self.params.runner_dict = defaultdict()

    def add_function(self, func, args=None, kwargs=None, key=None):
        """ Adds function in the Loom
        Args:
            func (reference): reference to the function
            args (list): function args
            kwargs (dict): function kwargs
            key (str): key to store the function output in dictionary
        """
        if args is None:
            args = list()
        if kwargs is None:
            kwargs = dict()
        if key is None:
            key = len(self.params.runners)
        self.params.runners.append((func, args, kwargs, key))
        
    def add_work(self, works):
        """ Adds work to the loom
        Args:
            works (list): list of works [(func, args, kwargs, key), (func2, args2, kwargs2), ...]
        """
        for work in works:
            if len(work) > 4 or len(work) == 0:
                raise ValueError('Need 1 to 4 values to unpack')
            args = work[1] if len(work) > 1 else None
            kwargs = work[2] if len(work) > 2 else None
            key = work[3] if len(work) == 4 else None
            self.add_function(work[0], args, kwargs, key)

    def execute(self):
        """ Executes runners and returns output dictionary containing runner output, error,
        started time, finished time and execution time of the given runner. runner key or the
        order in which function was added is the key to get the tracker dictionary.
        Returns:
            dict: output dict
            Examples:
                {
                    "runner1 key/order" : {
                            "output": runner output,
                            "error": runner errors,
                            "started_time": datetime.now() time stamp when runner was started
                            "finished_time": datetime.now() time stamp when runner was completed,
                            "execution_time: total execution time in seconds",
                            "got_error": boolean
                        },
                    "runner2 key/order" : {
                            "output": runner output,
                            "error": runner errors,
                            "started_time": datetime.now() time stamp when runner was started
                            "finished_time": datetime.now() time stamp when runner completed,
                            "execution_time: total execution time in seconds",
                            "got_error": boolean
                        }
                }
        """
        while self.params.runners:
            runner = self.params.runners.pop(0)
            self._start(runner)
            self.params.started.append(runner)
            while self._get_active_runner_count() >= self.params.max_runner:
                time.sleep(self.params.time_pause)
        while self._get_active_runner_count():
            time.sleep(self.params.time_pause)
        output = self.params.tracker_dict
        self.params.tracker_dict = defaultdict()
        self.params.runner_dict = defaultdict()
        return output
    
    def _get_active_runner_count(self):
        """ Returns the total number of runners running at the present time """
        count = 0
        for runner in self.params.started:
            if (self.params.runner_dict[runner[3]] and self.params.runner_dict[runner[3]].is_alive()
                or not self.params.tracker_dict[runner[3]]):
                    count += 1
        if count == 0:
            self.params.started = list()
        return count
    
    def _run(self, runner):
        """ Runs function runner """
        output, error, got_error = None, None, False
        started = datetime.datetime.now()
        try:
            output = runner[0](*runner[1], **runner[2])
        except Exception as e:
            got_error = True
            error = str(e)
        finally:
            finished = datetime.datetime.now()
            self.params.tracker_dict[runner[3]] = {
                "output": output,
                "started_time": started,
                "finished_time": finished,
                "execution_time": (finished - started).total_seconds(),
                "got_error": got_error,
                "error": error}
    
    def _start(self, runner):
        """ Starts runner process """
        self.params.runner_dict[runner[3]] = threading.Thread(target=self._run, args=(runner,))
        if self.params.mode:
            self.params.runner_dict[runner[3]].setDaemon(True)
        self.params.runner_dict[runner[3]].start()
        if not self.params.mode:
            self.params.runner_dict[runner[3]].join()
    



import numpy as np
from linora.utils._config import Config

__all__ = ['LRStep', 'LRConstant', 'LRCyclic', 'LRSGDR', 'LRStepMulti', 'LRScheduler', 'LRReduceOnPlateau']


class LRCyclic():
    """
    https://www.kaggle.com/hireme/fun-api-keras-f1-metric-cyclical-learning-rate/code
    
    This callback implements a cyclical learning rate policy (CLR).
    The method cycles the learning rate between two boundaries with
    some constant frequency, as detailed in this paper (https://arxiv.org/abs/1506.01186).
    The amplitude of the cycle can be scaled on a per-iteration or 
    per-cycle basis.
    This class has three built-in policies, as put forth in the paper.
    "triangular":
        A basic triangular cycle w/ no amplitude scaling.
    "triangular2":
        A basic triangular cycle that scales initial amplitude by half each cycle.
    "exp_range":
        A cycle that scales initial amplitude by gamma**(cycle iterations) at each 
        cycle iteration.
    For more detail, please see paper.
    
    Args:
        lr_min: initial learning rate which is the
            lower boundary in the cycle.
        lr_max: upper boundary in the cycle. Functionally,
            it defines the cycle amplitude (lr_max - lr_min).
            The lr at any cycle is the sum of lr_min
            and some scaling of the amplitude; therefore 
            lr_max may not actually be reached depending on
            scaling function.
        step_size: number of training iterations per
            half cycle. Authors suggest setting step_size
            2-8 x training iterations in epoch.
        mode: one of {triangular, triangular2, exp_range}.
            Default 'triangular'.
            Values correspond to policies detailed above.
            If scale_fn is not None, this argument is ignored.
        gamma: constant in 'exp_range' scaling function:
            gamma**(cycle iterations)
        scale_fn: Custom scaling policy defined by a single
            argument lambda function, where 
            0 <= scale_fn(x) <= 1 for all x >= 0.
            mode paramater is ignored 
        scale_mode: {'cycle', 'iterations'}.
            Defines whether scale_fn is evaluated on 
            cycle number or cycle iterations (training
            iterations since start of cycle). Default is 'cycle'.
    """
    def __init__(self, lr_min=0.001, lr_max=0.006, step_size=2000., mode='triangular',
                 gamma=1., scale_fn=None, scale_mode='cycle'):
        self._params = Config()
        self._params.lr = lr_min
        self._params.lr_min = lr_min
        self._params.lr_diff = lr_max-lr_min
        self._params.step_size = step_size
        self._params.mode = mode
        self._params.gamma = gamma
        if scale_fn is None:
            if self._params.mode == 'triangular':
                self._params.scale_fn = lambda x: 1.
                self._params.scale_mode = 'cycle'
            elif self._params.mode == 'triangular2':
                self._params.scale_fn = lambda x: 1/(2.**(x-1))
                self._params.scale_mode = 'cycle'
            elif self._params.mode == 'exp_range':
                self._params.scale_fn = lambda x: gamma**(x)
                self._params.scale_mode = 'iterations'
        else:
            self._params.scale_fn = scale_fn
            self._params.scale_mode = scale_mode
        self._params.name = 'LRCyclic'
        
    def _update(self, batch, log):
        """update log.
        
        Args:
            batch: Integer, index of batch.
            log: dict, name and value of loss or metrics;
        """
        cycle = np.floor(1+batch/(2*self._params.step_size))
        x = np.abs(batch/self._params.step_size - 2*cycle + 1)
        if self._params.scale_mode == 'cycle':
            self._params.lr = self._params.lr_min + self._params.lr_diff*np.maximum(0, (1-x))*self.scale_fn(cycle)
        else:
            self._params.lr = self._params.lr_min + self._params.lr_diff*np.maximum(0, (1-x))*self.scale_fn(batch)


class LRConstant():
    """Decays the learning rate of each parameter group by a small constant 
    factor until the number of batch reaches a pre-defined batch. 
    
    >>> # Assuming optimizer uses lr = 0.05
    >>> # lr = 0.025   if batch == 0
    >>> # lr = 0.025   if batch == 1
    >>> # lr = 0.025   if batch == 2
    >>> # lr = 0.025   if batch == 3
    >>> # lr = 0.05    if batch >= 4

    Args:
        lr_initial: lr initial value.
        factor: The number we multiply learning rate until the milestone.
        batch: The number of steps that the scheduler decays the learning rate. 
    """
    def __init__(self, lr_initial, factor, batch):
        self._params = Config()
        self._params.lr = lr_initial*factor
        self._params.factor = lr_initial
        self._params.batch = batch
        self._params.name = 'LRConstant'
    
    def _update(self, batch, log):
        """update log.
        
        Args:
            batch: Integer, index of batch.
            log: dict, name and value of loss or metrics;
        """
        if self._params.batch<=batch:
            self._params.lr = self._params.factor


class LRReduceOnPlateau():
    """Reduce learning rate when a metric has stopped improving.

    Args:
        lr_initial: lr initial value.
        scheduler: a function that takes current learning rate (float)  
            and an batch index (integer, indexed from 0) and log (dict) 
            as inputs and returns a new learning rate as output (float).
        monitor: quantity to be monitored.
        patience: number of batch with no improvement after which learning rate will be reduced.
        mode: one of {'min', 'max'}. 
            In 'min' mode, the learning rate will be reduced when the quantity monitored has stopped decreasing; 
            in 'max' mode it will be reduced when the quantity monitored has stopped increasing; 
        min_delta: Minimum change in the monitored quantity to qualify as an improvement, 
            i.e. an absolute change of less than min_delta, will count as no improvement.
        lr_min: lower bound on the learning rate.
    """
    def __init__(self, lr_initial, scheduler, monitor, patience=10, mode='min', min_delta=0.0001, lr_min=0):
        self._params = Config()
        self._params.lr = lr_initial
        self._params.scheduler = scheduler
        self._params.monitor = monitor
        self._params.patience = patience
        self._params.mode = mode
        self._params.min_delta = min_delta
        self._params.lr_min = lr_min
        self._params.history = []
        self._params.name = 'LRReduceOnPlateau'
    
    def _update(self, batch, log):
        """update log.
        
        Args:
            batch: Integer, index of batch.
            log: dict, name and value of loss or metrics;
        """
        if self._params.monitor in log:
            self._params.history = self._params.history[-self._params.patience:]+[log[self._params.monitor]]
            if self._params.mode=='min':
                if min(self._params.history[-self._params.patience:])+self._params.min_delta>self._params.history[0]:
                    self._params.lr = max(self._params.lr_min, self._params.scheduler(self._params.lr, batch, log))
            else:
                if max(self._params.history[-self._params.patience:])-self._params.min_delta<self._params.history[0]:
                    self._params.lr = max(self._params.lr_min, self._params.scheduler(self._params.lr, batch, log))


class LRScheduler():
    """Custom learning rate scheduler.

    Args:
        lr_initial: lr initial value.
        scheduler: a function that takes current learning rate (float)  
            and an batch index (integer, indexed from 0) and log (dict) 
            as inputs and returns a new learning rate as output (float).
    """
    def __init__(self, lr_initial, scheduler):
        self._params = Config()
        self._params.lr = lr_initial
        self._params.scheduler = scheduler
        self._params.name = 'LRScheduler'
    
    def _update(self, batch, log):
        """update log.
        
        Args:
            batch: Integer, index of batch.
            log: dict, name and value of loss or metrics;
        """
        self._params.lr = self._params.scheduler(self._params.lr, batch, log)


class LRSGDR():
    """Cosine annealing learning rate scheduler with periodic restarts.

    Args:
        min_lr: The lower bound of the learning rate range for the experiment.
        max_lr: The upper bound of the learning rate range for the experiment.
        steps_per_epoch: The number of batches per epoch
        lr_decay: Reduce the max_lr after the completion of each cycle.
                  Ex. To reduce the max_lr by 20% after each cycle, set this value to 0.8.
        cycle_length: Initial number of epochs in a cycle.
        mult_factor: Scale epochs_to_restart after each full cycle completion.
        
    References
        Original paper: http://arxiv.org/abs/1608.03983
    """
    def __init__(self, lr_min, lr_max, steps_per_epoch, lr_decay=1, cycle_length=10, mult_factor=2):
        self._params = Config()
        self._params.lr = lr_max
        self._params.lr_min = lr_min
        self._params.lr_max = lr_max
        
        self._params.lr_decay = lr_decay

        self._params.batch_since_restart = 0
        self._params.next_restart = cycle_length

        self._params.steps_per_epoch = steps_per_epoch

        self._params.cycle_length = cycle_length
        self._params.mult_factor = mult_factor
        self._params.name = 'LRSGDR'

    def _update(self, batch, log):
        """update log.
        
        Args:
            batch: Integer, index of batch.
            log: dict, name and value of loss or metrics;
        """
        self._params.batch_since_restart += 1
        restart = self._params.batch_since_restart / (self._params.steps_per_epoch * self._params.cycle_length)
        self._params.lr = self._params.lr_min+0.5*(self._params.lr_max-self._params.lr_min)*(1+np.cos(restart*np.pi))

        if np.ceil(batch/self._params.steps_per_epoch) + 1 == self._params.next_restart:
            self._params.batch_since_restart = 0
            self._params.cycle_length = np.ceil(self._params.cycle_length * self._params.mult_factor)
            self._params.next_restart += self._params.cycle_length
            self._params.lr_max *= self._params.lr_decay


class LRStep():
    """Decays the learning rate of each parameter group by gamma every step_size batch. 
    
    >>> # Assuming optimizer uses lr = 0.05 
    >>> # lr = 0.05     if batch < 30
    >>> # lr = 0.005    if 30 <= batch < 60
    >>> # lr = 0.0005   if 60 <= batch < 90

    Args:
        lr_initial: lr initial value.
        step_size: int, Period of learning rate decay.
        gamma: float, Multiplicative factor of learning rate decay. Default: 0.1.
    """
    def __init__(self, lr_initial, step_size, gamma=0.1):
        self._params = Config()
        self._params.lr = lr_initial
        self._params.step_size = step_size
        self._params.gamma = gamma
        self._params.name = 'LRStep'
        self._params.step_num = 0
        self._params.batch = -1
    
    def _update(self, batch, log):
        """update log.
        
        Args:
            batch: Integer, index of batch.
            log: dict, name and value of loss or metrics;
        """
        if self._params.batch!=batch:
            self._params.batch = batch
            self._params.step_num += 1
            if self._params.step_num%self._params.step_size==0:
                self._params.lr = self._params.lr*self._params.gamma


class LRStepMulti():
    """Decays the learning rate of each parameter group by 
    gamma once the number of batch reaches one of the batch_list.  
    
    >>> # Assuming optimizer uses lr = 0.05
    >>> # lr = 0.05     if batch < 30
    >>> # lr = 0.005    if 30 <= batch < 80
    >>> # lr = 0.0005   if batch >= 80

    Args:
        lr_initial: lr initial value.
        batch_list: list, List of batch indices.
        gamma: float, Multiplicative factor of learning rate decay. Default: 0.1.
    """
    def __init__(self, lr_initial, batch_list, gamma=0.1):
        self._params = Config()
        self._params.lr = lr_initial
        self._params.batch_list = sorted(batch_list)
        self._params.gamma = gamma
        self._params.name = 'LRStepMulti'
    
    def _update(self, batch, log):
        """update log.
        
        Args:
            batch: Integer, index of batch.
            log: dict, name and value of loss or metrics;
        """
        for r, i in enumerate(self._params.batch_list.copy()):
            if batch>i:
                self._params.lr = self._params.lr*self._params.gamma
                self._params.batch_list.pop(r)

